#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//

try:
    from opensees.openseespy import Model as _Model
except:
    Model = object

successful = 0


from ._analysis import StaticAnalysis, EigenAnalysis, DynamicAnalysis
from ._material import Material
from ._section import Section, ShellSection
from .load import NodalLoad


try:
    from jax import tree_util
#   import opensees.openseespy as ops
# 1) Tell JAX that ops.Model is an opaque leaf
    def _flatten_model(model):
        # no children, carry the model object in aux_data
        return (), model

    def _unflatten_model(aux_data, _children):
        # aux_data *is* the original model
        return aux_data

    tree_util.register_pytree_node(Model, _flatten_model, _unflatten_model)
except:
    pass


class Model(_Model):
    pass
    # @property 
    # def nodes(self):
    #     nn = self.getNumNodes()


def solve(model: Model,
          loading,
          output=None,
          analysis=None,
          algorithm=None, scale=1.0,
          test=None):
    """
    Solve a static equilibrium problem for the model with the given loading.

    Parameters
    ----------
    model : :py:class:`Model`
        The structural model to analyze.
    loading : Load or list of Loads
        The loads to apply during the analysis.
    algorithm : Algorithm, optional
        The solution algorithm to use. If None, the :ref:`Newton` algorithm is used.
    constraints : ConstraintHandler, optional
        The constraint handler to manage boundary conditions. Default is :ref:`"Plain" <PlainConstraints>`.
    """
    analysis_ = StaticAnalysis(loading, algorithm=algorithm, test=test)
    if analysis_.analyze(model) != successful:
        message = "Analysis did not complete successfully."
        message = model._openseespy._interp.read_error()
        raise RuntimeError(message)

    if output is not None:
        node, dof = output
        return model.nodeDisp(node, dof)

    return model.state


def eigen(model, modes, solver=None, output=None, problem="Frequency"):
    """
    Perform eigenvalue analysis on the model.

    Parameters
    ----------
    model : :py:class:`Model`
        The structural model to analyze.
    modes : int
        The number of eigenmodes to compute.
    solver : EigenSolver, optional
        The eigenvalue solver to use. If None, a default solver is used.
    problem : str, optional
        The type of eigenvalue problem to solve. Options are "Frequency" or "Buckling".

    Returns
    -------
    eigenvalues : list
        The computed eigenvalues.
    """
    model.eigen(modes)


def trace(model, loading, output=None, algorithm=None, method=None):
    """
    Perform a static analysis while tracing the response of the model.
    """
    analysis = StaticAnalysis(loading, algorithm=algorithm)
    if analysis.analyze(model) != successful:
        raise RuntimeError("Analysis did not complete successfully.")

    if output is not None:
        node, dof = output
        return model.nodeDisp(node, dof)
    return None



def integrate(model, loading, dt, steps, output=None, algorithm=None, method=None):
    """
    integrate the dynamic response of the model.
    """
    if output is None:
        output = {"displacement": model.getNodeTags()}
        output["velocity"] = output["displacement"]
        output["acceleration"] = output["displacement"]

    analysis = DynamicAnalysis(loading, dt, algorithm=algorithm)

    analysis._start_equilibrium(model)
    from collections import defaultdict
    result = defaultdict(list)
    for step in range(steps):
        if analysis.analyze(model, dt, 1) != successful:
            message = model._openseespy._interp.read_error()
            raise RuntimeError(f"Analysis did not complete successfully: {message}")
        if output is not None:
            for key, nodes in output.items():
                if key == "displacement":
                    for node in nodes:
                        result["U"].append(model.nodeDisp(node))
                elif key == "acceleration":
                    for node in nodes:
                        result["A"].append(model.nodeAccel(node))
                elif key == "velocity":
                    for node in nodes:
                        result["V"].append(model.nodeVel(node))
    return dict(result)


class Integrator:
    pass

class _LoadControl:
    pass

class ArcLengthControl(_LoadControl):
    pass

class DisplacementControl(_LoadControl):
    pass

class EnergyControl(_LoadControl):
    pass

class LoadControl(_LoadControl):
    pass



import numpy as np
from typing import Dict, Any, Callable
from .convergence import PlotConvergenceRate


class CellField(Callable): ...

class NodeField(Callable): ...

class PatchRecovery:
    pass 

class MomentDiagram:
    pass

class Curvature:
    pass

class Resultants:
    pass

class FrameMoments(CellField):
    def __init__(self, model,  artist):
        self._model = model
        self._artist = artist

    def __call__(self, tag):
        ndm = self._artist.model.ndm
        wy  = -10 if tag == 2 else (-5 if tag == 1 else 0)
        wz  =  -1 if tag == 1 else 0

        X = self._artist.model.cell_position(tag)
        L = np.linalg.norm(X[-1] - X[0])
        if self._artist.model.cell_matches(tag, "prism"):
            xe = np.linspace(0, 1, 20)
            q = self._model.eleResponse(tag, "basicForces")


            if ndm == 3:
                w  = np.array([   0 ,  wz ,  wy ])
                qi = np.array([-q[3], q[4], q[1]])
                qj = np.array([ q[3], q[5], q[2]])
            else:
                w  = np.array([0, 0,  wy ])
                qi = np.array([0, 0, q[1]])
                qj = np.array([0, 0, q[2]])

            ye = [ (qi*(x-1) + qj*x + (1-x)*x * w * L*L/2) for x in xe ]
            return xe, ye

        else:

            # xe = [x for x,_ in self._artist.model.cell_quadrature(tag)]
            xe = [x/L for x in self._model.eleResponse(tag, "integrationPoints")]
            ye = [self._model.eleResponse(tag, "section", i+1, "deformations")[3:6]
                  for i in range(len(xe))]
            return xe, ye


class NodalAverage(NodeField):
    def __init__(self, model, response: CellField, form=None, component=None, method=None):
        """
        Return a dictionary of stress components per node.
        For 2D: sxx, syy, sxy
        For 3D: sxx, syy, szz, sxy, syz, sxz
        Summed values from each element's node contribution are averaged at the end.

        Args:
            model: an OpenSees model-like object with the following interface:
                - getEleTags()
                - getEleClassTags(tag: int)
                - getNodeTags()
                - eleNodes(tag: int)
                - eleResponse(tag: int, key: str)
            response (str): a label if needed for e.g. 'stress', 'stressAtNodes'
            ndm (int): number of dimensions (2 or 3)

        Returns:
            dict[node_tag] = {
                "xx": float,
                "yy": float,
                ...
            }
        """
        self._component = component
        if component is None:
            if form is not None:
                raise ValueError("If 'component' is None, 'form' must also be None.")
            form = "scalar"
        elif form is None:
            ndm = model.getNDM()
            if ndm == 2:
                form = "plane"
            elif ndm == 3:
                form = "solid"
        self._form = form

        if isinstance(response, str):
            self._resp_func = lambda elem: model.eleResponse(elem, response)
        elif callable(response):
            self._resp_func = response
        else:
            raise ValueError("response must be a string or callable")

        if form == "frame":
            keys = "sxx", "sxy", "sxz"
        elif form == "plane":
            keys = "sxx", "syy", "sxy"
        elif form == "solid":
            keys = "sxx", "syy", "szz", "sxy", "syz", "sxz"
        elif form == "scalar":
            keys = [None]
        self._keys = keys


        try: #if hasattr(model, "cells"):
            output = self._call_meshio(model)

        except Exception as e:
            raise e #else:
            output = self._call_xara(model)


        # Now do averaging (if a node belongs to multiple elements).
        for node in output:
            c = output[node]["_count"] or 1.0

            for key in keys:
                output[node][key] /= c

        # Clean up
        for node in output:
            output[node].pop("_count", None)


        self._values = output


    def _call_meshio(self, model):
        keys = self._keys
        nrc = len(keys)

        output = {
            node: {
                "_count": 0,
                **{key: 0.0 for key in keys}
            } for node in range(len(model.nodes))
        }

        # Assume each element’s response yields a flat array 
        # with length = nen * nrc.
        # Then we reshape it to (nen, nrc).
        for elem, elem_nodes in enumerate(model.cells()):
            nen = len(elem_nodes)

            se = self._resp_func(elem)
            # try:
            #     nr = len(se)
            # except TypeError:
            #     if self._form != "scalar":
            #         raise
            #     se = [se]

            assert not (len(se) % nen), f"Element {elem} response length mismatch: {len(se)} != {nen * nrc}"
            se = np.reshape(se, (nen, nrc))

            # Accumulate at element's nodes
            for i, node in enumerate(elem_nodes):
                for j, key in enumerate(keys):
                    output[node][key] += se[i][j]

                output[node]["_count"] += 1

        return output

    def _call_xara(self, model):
        keys = self._keys

        nrc = len(keys)
        output = {
            node: {
                "_count": 0,
                **{key: 0.0 for key in keys}
            } for node in model.getNodeTags()
        }

        # Assume each element’s response yields a flat array 
        # with length = nen * nrc.
        # Then reshape it to (nen, nrc).
        for elem in model.getEleTags():
            elem_nodes = model.eleNodes(elem)
            nen = len(elem_nodes)

            se = self._resp_func(elem)
            assert not (len(se) % nen), f"Element {elem} response length mismatch: {len(se)} != {nen * nrc}"
            se = np.reshape(se, (nen, nrc))

            # Accumulate at element's nodes
            for i, node in enumerate(elem_nodes):
                for j, key in enumerate(keys):
                    output[node][key] += se[i][j]

                output[node]["_count"] += 1
        return output

    def __call__(self, node):
        s = self._values[node]
        if self._component is None and self._form != "scalar":
            return s

        elif self._component in s:
            return s[self._component]

        elif self._component.lower() in {"j2", "svm"}:
            if self._form == "frame":
                P = np.array([[1, 0, 0], [0, 3, 0], [0, 0, 3]], dtype=float)
                s = self._values[node]
                s = s["sxx"], s["sxy"], s["sxz"]
                return np.sqrt(np.dot(s, P@s))
            else:
                return np.sqrt(
                    (s["sxx"] - s["syy"])**2 +
                    (s["syy"] - s["szz"])**2 +
                    (s.get("szz",0) - s["sxx"])**2 +
                    6*(s["sxy"]**2 + s.get("syz",0)**2 + s.get("sxz",0)**2)
                )/np.sqrt(2)



def FiberStress(model,
                 shape,
                 section: int,
                 stress : str,
                 element: int=1):

    nen = 3
    def f(fiber):
        s = model.eleResponse(element, "section", section, "fiber", fiber, "stress")
        return [s]*nen

    return NodalAverage(shape.model, f, component=stress, form="frame")


class NodalStress(NodalAverage):
    def __init__(self, model, component=None, method=None, form=None):
        import xara
        if not isinstance(model, xara.Model):
            model = model._model 
        

        ndm = 2
        if component is None:
            if form is not None:
                raise ValueError("If 'component' is None, 'form' must also be None.")
            form = "scalar"
        elif form is None:
            if ndm == 2:
                form = "plane"
            elif ndm == 3:
                form = "solid"

        self._model = model
        self._component = component

        super().__init__(model, "stressAtNodes",
                         component=component,
                         method=method,
                         form=form)


def node_average(model, response: CellField, ndm=2, keys=None):
    """
    Return a dictionary of stress components per node.
    For 2D: sxx, syy, sxy
    For 3D: sxx, syy, szz, sxy, syz, sxz
    Summed values from each element's node contribution are averaged at the end.

    Args:
        model: an OpenSees model-like object with the following interface:
               - getEleTags()
               - getEleClassTags(tag: int)
               - getNodeTags()
               - eleNodes(tag: int)
               - eleResponse(tag: int, key: str)
        response (str): a label if needed for e.g. 'stress', 'stressAtNodes'
        ndm (int): number of dimensions (2 or 3)

    Returns:
        dict[node_tag] = {
            "xx": float,
            "yy": float,
            ...
        }
    """
    if isinstance(response, str):
        resp_func = lambda elem: model.eleResponse(elem, response)
    elif callable(response):
        resp_func = response
    else:
        raise ValueError("response must be a string or callable")


    # For a typical 2D element with 3 stress components: sxx, syy, sxy
    # For a typical 3D element with 6 stress components: sxx, syy, szz, sxy, syz, sxz
    # You may need to adapt this depending on your element type(s).
    if ndm == 2:
        keys = "sxx", "syy", "sxy"
    else:
        keys = "sxx", "syy", "szz", "sxy", "syz", "sxz"

    nrc = len(keys)

    # Initialize a dictionary to hold stress sums and a count of how many
    # times each node is encountered (to do averaging).
    output = {node: {"_count": 0, **{key: 0.0 for key in keys}} for node in model.getNodeTags()}


    # We assume each element’s response yields a flat array 
    # with length = nen * nrc.
    # Then we reshape it to (nen, nrc).
    for elem in model.getEleTags():
        elem_nodes = model.eleNodes(elem)
        nen = len(elem_nodes)

        se = resp_func(elem)
        assert not (len(se) % nen),  f"Element {elem} response length mismatch: {len(se)} != {nen * nrc}"
        se = np.reshape(se, (nen, nrc))

        # Accumulate at element's nodes
        for i, node in enumerate(elem_nodes):
            for j, key in enumerate(keys):
                output[node][key] += se[i][j]

            output[node]["_count"] += 1


    # Now do averaging (if a node belongs to multiple elements).
    for node in output:
        c = output[node]["_count"] or 1.0

        for key in keys:
            output[node][key] /= c


    # Clean up
    for node in output:
        output[node].pop("_count", None)

    return output


def _stress_spr(xyz,  ElemData, Stress) -> Dict[str, np.ndarray]:
    """
    Super-patch stress recovery for quadrilateral membrane elements.

    - Zienkiewicz and Zhu (1992) The superconvergent patch recovery and a posteriori error estimates. 
      Part 1: The recovery technique DOI: https://doi.org/10.1002/nme.1620330702


    Parameters
    ----------
    xyz : (ndm, nen) ndarray
        Undeformed nodal coordinates.
    ElemState : (something) ndarray
        Element-level displacement array (layout compatible with `ExtrReshu`).
    ElemData : object
        Element-specific data structure.  Must expose at least:
            • nIP   – number of Gauss points per direction
            • nodix – local-to-global node index map
            • Geom  – geometric-transformation data structure
    Stress : callable
        Material routine.  Must accept the signature
        `Stress('post', ip_index)` and return an object
    ndm : int, default 2
        Spatial dimension (2 ⇒ membrane problems).
    """
    nen  = xyz.shape[1]            # number of nodes

    # Geometric transformations ---------------------------------------
    xyzl, _   = DefGeom_Quad(xyz)                         # local xyz

    # ------------------------------------------------------------------
    # Gauss integration
    # ------------------------------------------------------------------
    nstr = 4                                # σ-vector length (σ_x, σ_y, τ_xy, p)
    sigNd = np.zeros((nen, nstr))
    LSQM  = np.zeros((nen, nen))

    # Gauss points & weights (nIP × nIP rule) --------------------------
    nat, wIP = Gauss2d(ElemData.nIP)        # nat.shape = (nIP**2, 2)

    for i, (ξ, η) in enumerate(nat):
        # Shape functions & derivatives -------------------------------
        N, _, J = shape2d((ξ, η), xyzl[:2, :].T, ElemData.nodix)

        sig = Stress("post", i)

        # Accumulate recovery terms -----------------------------------
        detJ = float(np.linalg.det(J))
        weight = detJ * wIP[i]

        sigNd += np.outer(N, sig) * weight
        LSQM  += np.outer(N, N) * weight

    return {"sig": sigNd, "LSQM": LSQM}


def project_element_stress_to_nodes(xyz, ElemData, Stress) -> Dict[str, np.ndarray]:
    """
    Project Gauss-point stresses to the element’s nodes with a least-squares fit.

    Parameters
    ----------
    xyz : (ndm, nen) ndarray
        Undeformed nodal coordinates (global system).
    ElemData : object
        Element-specific data structure that exposes
            • nIP   – Gauss-rule order (nIP × nIP)
            • nodix – local-to-global node map (passed straight to `shape2d`)
    Stress : callable
        A routine that returns the (already *global*) stress vector at an
        integration point: `sig = Stress("post", ip_index)`
        `sig` must be length-4 (σₓ, σ_y, τ_xy, p).
    ndm : int, default 2
        Spatial dimension (membranes ⇒ 2).

    """
    nen = xyz.shape[1]                         # # nodes

    # local geometry (no displacement-based rotation needed here)
    xyzl, _ = DefGeom_Quad(xyz)                # (3, nen)

    # integration rule -------------------------------------------------
    nat, wIP = Gauss2d(ElemData.nIP)           # nat: (nIP**2, 2)

    r    = np.zeros((nen, 4))                  # RHS vector
    LSQM = np.zeros((nen, nen))                # least-squares matrix

    # ------------------------------------------------------------------
    # loop over Gauss points
    # ------------------------------------------------------------------
    for i, (xi, eta) in enumerate(nat):
        # shape functions & Jacobian -----------------------------------
        N, _dNdx, J = shape2d((xi, eta), xyzl[:2, :].T, ElemData.nodix)

        sig = Stress("post", i)                # (4,) stress in *global* coords

        detJ = float(np.linalg.det(J))
        w    = detJ * wIP[i]

        r    += np.outer(N, sig) * w
        LSQM += np.outer(N, N) * w

    # ------------------------------------------------------------------
    # nodal stresses: solve LSQM · sigNd = r
    # ------------------------------------------------------------------
    # (nen × nen) · (nen × 4) = (nen × 4)
    try:
        sigNd = np.linalg.solve(LSQM, r)
    except np.linalg.LinAlgError:
        # fall back to least-squares if LSQM is singular/ill-conditioned
        sigNd = np.linalg.lstsq(LSQM, r, rcond=None)[0]

    return {"sigNd": sigNd, "LSQM": LSQM}




class Domain:
    def __init__(self, partitions):
        from multiprocessing import Process, Queue

        self.num_parts = len(partitions)
        self.workers = []
        self.sub_messages = []
        self.sub_responses = []
        self.interface_guess = {}  # {(node, dof): value} for Dirichlet

        for part in partitions:
            message = Queue()
            response = Queue()

            # Start a new child process which will construct a model.
            p = Process(target=self._subdomain_worker, args=(part, message, response))
            p.start()

            self.workers.append(p)
            self.sub_messages.append(message)
            self.sub_responses.append(response)

    def _subdomain_worker(self, partition, messageueue, responseueue):
        sd = Subdomain(partition)
        while True:
            msg = messageueue.get()
            if msg == "TERMINATE":
                break

            elif msg["cmd"] == "solve":
                data = msg["data"]
                result = sd.solve_local(data)
                responseueue.put(result)

            elif msg["cmd"] == "schur":
                S, g = sd.get_schur_data()
                responseueue.put((S, g))

    def step(self):
        """
        Perform a single iteration:
        - Send current interface Dirichlet guess to all subdomains
        - Collect reactions
        - Return aggregate response (could be used to update guesses)
        """
        interface_data = {
            "dirichlet": self.interface_guess.copy(),
            "neumann": {}  # Could also support this later
        }

        # Send solve command with interface data to each subdomain
        for message in self.sub_messages:
            message.put({"cmd": "solve", "data": interface_data})

        # Collect responses
        all_reactions = []
        for response in self.sub_responses:
            result = response.get()
            reactions = result.get("reactions", {})
            all_reactions.append(reactions)

        # Optional: aggregate or store reactions for update
        return all_reactions

    def shutdown(self):
        for q in self.sub_messages:
            q.put("TERMINATE")
        for p in self.workers:
            p.join()

    def schur_update(self):
        S_total = None
        g_total = None

        # Request Schur data from all subdomains
        for message in self.sub_messages:
            message.put({"cmd": "schur"})

        for response in self.sub_responses:
            S_i, g_i = response.get()
            S_total = S_i if S_total is None else S_total + S_i
            g_total = g_i if g_total is None else g_total + g_i

        # Solve interface system
        u_gamma = np.linalg.solve(S_total, g_total)

        # Update interface guess
        # Map back into {(node,dof): value}
        self.interface_guess = self.unpack_interface_vector(u_gamma)




class Subdomain:
    def __init__(self, partition):
        import xara
#       self.model = xara.Model()
        import openseespy.opensees as ops

        model = ops
        self.partition = partition
        partition.populate(model)

        dofs = partition.get_dof_partition()
        self.interior_dofs  = dofs["interior"]
        self.interface_dofs = dofs["interface"]

    def solve_local(self, interface_data: dict):
        self.apply_interface_conditions(interface_data)
        self.model.analyze(1)
        return self.extract_interface_response()

    def apply_interface_conditions(self, interface_data: dict):
        dirichlet = interface_data.get("dirichlet", {})
        neumann = interface_data.get("neumann", {})

        for (node, dof), value in dirichlet.items():
            self.model.remove('sp', node, dof)  # Ensure clean slate
            self.model.fix(node, dof, value)

        for (node, dof), value in neumann.items():
            self.model.load(node, *self._vector_for_dof(dof, value))

    def extract_interface_response(self) -> dict:
        """Return reaction forces on interface DOFs."""
        responses = {"reactions": {}}
        for node in self.interface_dofs:
            for dof in self._active_dof_dirs(node):
                key = (node, dof)
                val = self.model.nodeReaction(node, dof)
                responses["reactions"][key] = val
        return responses

    def _vector_for_dof(self, dof: int, val: float):
        """Return a nodal load vector with `val` in the `dof`th position."""
        # Assuming 3 DOFs per node (can adapt)
        v = [0.0, 0.0, 0.0]
        v[dof-1] = val
        return v

    def _active_dof_dirs(self, node: int):
        # TODO: Replace with real DOF info from model or partition
        return [1, 2, 3]  # Assuming 3 translational DOFs

    def get_schur_data(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Return (S_i, g_i) for the interface DOFs.
        Assumes local tangent has been assembled.
        """
        # TODO(cmp) This should be done directly in OpenSees
        K = self.model.getTangent()
        R = self.model.getResidual()

        # Partition DOFs
        I = self.interior_dofs
        Γ = self.interface_dofs

        # Extract subblocks of K (using DOF maps)
        K_II = K[np.ix_(I, I)]
        K_IΓ = K[np.ix_(I, Γ)]
        K_ΓI = K_IΓ.T
        K_ΓΓ = K[np.ix_(Γ, Γ)]

        # Compute local Schur complement
        S = K_ΓΓ - K_ΓI @ np.linalg.solve(K_II, K_IΓ)

        # Compute local RHS
        f_I = R[I]
        f_Γ = R[Γ]
        g = f_Γ - K_ΓI @ np.linalg.solve(K_II, f_I)

        return S, g






#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//

class Newton:
    """Newton()
    Use the Newton-Raphson method for solving nonlinear equations.
    """
    def __init__(self):
        pass

    def _setup(self, model):
        model.algorithm("Newton")


class NewtonLineSearch:
    """
    Use the Newton-Raphson method with line search for solving nonlinear equations.
    
    Parameters
    ----------
    tolerance : float
        The line search tolerance. Default is 0.8.
    """
    def __init__(self, tolerance=0.8):
        self.tolerance = tolerance

    def _setup(self, model):
        model.algorithm("NewtonLineSearch", self.tolerance)

class ModifiedNewton:
    def __init__(self, initial=False):
        self.initial = initial

    def _setup(self, model):
        if self.initial:
            model.algorithm("ModifiedNewton", initial=True)
        else:
            model.algorithm("ModifiedNewton")


class BFGS:
    def __init__(self):
        pass

    def _setup(self, model):
        model.algorithm("BFGS")




OPTIONS = {
    "storage": ["Full", "Band", "Sparse", "Profile"],
}

SOLVERS = {
    "ProfileSPD": {"symmetric": [True],  "storage": "Profile", "driver": "ops_profile"},
    "Umfpack":    {"symmetric": [False], "storage": "Sparse",  "driver": "umfpack"},
    "Mumps":      {"symmetric": [False], "storage": "Sparse",  "driver": "mumps"},
    "SparseSPD":  {"symmetric": [True],  "storage": "Sparse",  "driver": "ops_sparse"},
    "BandGen":    {"symmetric": [False], "storage": "Band",    "driver": "DGBSV"},
    "BandSPD":    {"symmetric": [True],  "storage": "Band",    "driver": "DPBSV"},
    "FullGen":    {"symmetric": [False], "storage": "Full",    "driver": "DGESV"},
    "FullSPD":    {"symmetric": [True],  "storage": "Full",    "driver": "DPOSV"},
}


class LinearSystem:
    """
    LinearSystem("Mumps", symmetric=False)
    LinearSystem(detect=model)
    """

    def __init__(self,
                 first:     str = None,
                 solver:    str = None,
                 storage:   str = None,
                 symmetric: bool = None,
                 definite:  bool = None,
                 numberer:  str = None):
        
        if isinstance(first, str):
            if first.lower() in ["mumps", "umfpack"]:
                storage = "Sparse"
            elif "band" in first.lower():
                storage = "Band"
            elif "full" in first.lower():
                storage = "Full"
            elif "profile" in first.lower():
                storage = "Profile"
            elif "sparse" in first.lower():
                storage = "Sparse"
            else:
                raise ValueError(f"Unknown solver '{first}'")
            name = first
        
        elif first is None:
            if symmetric is True and storage in ["Full", "Band"]:
                name = "FullSPD" if storage == "Full" else "BandSPD"
            elif symmetric is True and storage == "Sparse":
                name = "SparseSPD"
            elif symmetric is False and storage in ["Full", "Band"]:
                name = "FullGen" if storage == "Full" else "BandGen"
            elif symmetric is False and storage == "Sparse":
                name = "Mumps"
            else:
                name = "Mumps"

        if symmetric is None:
            if name.lower() in ["mumps", "umfpack"]:
                symmetric = False
            else:
                symmetric = True


        if numberer is None:
            if storage.lower() in ["band", "profile"]:
                numberer = "RCM"
            elif storage.lower() in ["sparse"]:
                numberer = "AMD"
            else:
                numberer = "Plain"

        self.name = name
        self.solver = solver
        self.storage = storage
        self.numberer = numberer
        self.symmetric = symmetric
        self.definite = definite

    def __repr__(self):
        return f"LinearSystem('{self.name}', symmetric={self.symmetric}, storage='{self.storage}', numberer='{self.numberer}', definite={self.definite})"


"""Constants for consistently converting between units.


FPS
fps(a) or ae
  (Absolute, or mass-based) Foot-Pound-Second System.

fps(g) or bg
  (Gravitational, or force-based) Foot-Pound-Second System. Also known as the
  technical FPS system or British Gravitational system

fps(e) or ee
  (weight-based) Foot-Pound-Second System. Also known as the English Engineering
  System.

cgs 
  Centimetre-Gram-Second System



See also:

- http://130.149.89.49:2080/v2016/books/popups/usb-tbl.html

"""

from .units import *
from . import fps, cgs, mks, iks, si, us





def invoke_uniaxial(name, *args, **kwds):
    from opensees.openseespy import Model
    model = Model(ndm=1, ndf=1)
    tag = 1
    model.uniaxialMaterial(name, tag, *args, **kwds)

    return _Handle(model._openseespy, "UniaxialMaterial", tag)

class _Handle:
    def __init__(self, interpreter, type, tag=None, *args):
        self._interpreter = interpreter
        self._type = type
        self._tag = tag

    # def __getattribute__(self, name):
    #     if name in ["getStress", "get_stress", "get_tangent"]:
    #         return partial(self._interpreter._invoke_proc, 
    #                    "invoke", self._type, self._tag, name)

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def set_strain(self, *args):
        pass

    def getStress(self, e, commit=False):
        self._interpreter._invoke_proc("invoke", self._type, self._tag, [f"strain {e};"])
        if commit:
            self._interpreter._invoke_proc("invoke", self._type, self._tag, "commit")

        return self._interpreter._invoke_proc("invoke", self._type, self._tag, "stress")



    def getTangent(self, *args):
        return self._interpreter._invoke_proc("invoke", self._type, self._tag, "tangent") or 0
    





class Node:
    def __init__(self, model, tag):
        self._model = model
        self._tag = tag

    @property 
    def location(self):
        """Get the location of the node."""
        return self._model.nodeCoord(self._tag)
    
    


#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//
import jax
import inspect
from functools import wraps


def NodalValue(create_model, output, steps=1):
    node, dof = output

    from xara.para import _Parameter

    pnames = list(inspect.signature(create_model).parameters.keys())
    @jax.custom_vjp
    @wraps(create_model)
    def f(*params):
        m = create_model(*(
            _Parameter(p, index=i, name=n) for (i, p), n in zip(enumerate(params), pnames)
        ))
        m.analyze(steps)
        return m.nodeDisp(node, dof)

    # Forward
    def f_fwd(*params):
        m = create_model(*(
            _Parameter(p, index=i, name=n) for (i, p), n in zip(enumerate(params), pnames)
        ))
        m.analyze(steps)
        y = m.nodeDisp(node, dof)
        return y, m


    # Backward
    def f_bwd(m, cotangent):
        return tuple(
            m.sensNodeDisp(node, dof, tag) * cotangent
            for tag in m._model._openseespy._interp._tcl.call("getParamTags")
        )

    f.defvjp(f_fwd, f_bwd)
    return f




from ._material import Material, _Material

class ShellSection:
    def __init__(self, type, material, shape, **kwds):
        if not isinstance(material, dict):
            material = {
                None: material
            }
        self.materials = material
        self.shape = shape
        self.kwds = kwds
        self.type = type

    def _add_to_model(self, model, tag):
        if "fiber" in self.type.lower():
            if isinstance(self.shape, float):
                nip = 5
                mtag = self.materials[None]._mtag
                if mtag is None:
                    raise ValueError("Material must be added to model before section.")
                thickness = self.shape
                fibers = [(mtag, thickness/nip) for _ in range(nip)]
                fibers = [x for pair in fibers for x in pair]
                model.section("LayeredShell", tag, nip, *fibers)
            else:
                raise NotImplementedError("Only uniform fiber sections are implemented.")
        else:
            raise NotImplementedError(f"Shell section type '{self.type}' is not implemented.")



def Section(type, shape, material=None, mixed_type=None, mixed=True, nu=None):
    if not isinstance(material, (dict, _Material,None.__class__)):
        shape, material = material, shape
    
    if material is None:
        material = shape.materials
    elif not isinstance(material, dict):
        material = {
            None: material
        }

    if type == "Elastic":
        return _ElasticSection(type, shape, material, mixed_type=mixed_type, mixed=mixed, nu=nu)
    elif type in {"ShearFiber", "Multiaxial"}:#, "AxialFiber"}:
        return _MultiaxialSection(type, shape, material, mixed_type=mixed_type, mixed=mixed, nu=nu)
    elif type in {"AxialFiber"}:
        return _UniaxialSection(type, shape, material, mixed_type=mixed_type, mixed=mixed, nu=nu)
    else:
        raise NotImplementedError(f"Section type '{type}' is not implemented.")




class _Section:
    def __init__(self, type, shape, material, mixed_type=None, nu=None, mixed=True):
        self.type = type
        self._tag = None

        self._shape = shape
        if material is None:
            material = shape.materials
        self._material = material

        if mixed_type == "energetic":
            mixed_type = "UE"
        elif mixed_type == "geometric":
            mixed_type = "UG"

        self.__mixed_type = mixed_type
        self._mixed = mixed

        from xsection.analysis import SaintVenantSectionAnalysis
        self._shape_model = shape.model
        sv = SaintVenantSectionAnalysis(shape, nu=nu, model=self._shape_model)

        self.GJ = sv.twist_rigidity()
        self.EA = sv.axial_rigidity()
        self.rc = sv.centroid()
        self.sv = sv


    def _mixed_type(self, model):
        if self.__mixed_type is not None:
            return self.__mixed_type
        
        ndm = model._call("getNDM")
        ndf = model._call("getNDF")
        if ndm == 2 or not self._mixed:
            return None 
        elif ndf == 6:
            return "UE"
        elif ndf == 7:
            return "NR"

        
    def _warp_shapes(self, model, mixed_type):
        shape = self._shape

        warp_shapes = []

        ndf = model._call("getNDF")

        if mixed_type in {"UE", "UG", "UT", "NR"} or ndf >= 7:
            sv = self.sv
            warp_shapes.append(sv.solve_twist())
            if mixed_type in {"UE", "UG"} or ndf >= 8:
                warp_shapes.extend(sv.solve_shear())
            elif mixed_type in {"NR"}:
                warp_shapes.append(sv.solve_mixed_twist())

        return warp_shapes

    @property
    def materials(self)->dict:
        return self._material

    def _add_to_model(self, model, tag):
        raise NotImplementedError("Subclasses must implement _add_to_model method.")


class _UniaxialSection(_Section):
    def __init__(self, type, shape, material, **kwds):
        super().__init__(type, shape, material, **kwds)


    def _add_to_model(self, model, tag):
        shape = self._shape

        mixed_type = self._mixed_type(model)

        sdata = {
            "mixed_type": mixed_type,
        }
        if mixed_type in {"UE", "UG", "UT", "NV", "NR", "NT"}:
            GJ = self.GJ
        elif mixed_type is None:
            GJ = 1e4 # TODO

        mixed_type = self._mixed_type(model)
        if False: #mixed_type in {"UE", "UG"}:
            trace = shape.create_trace(form=mixed_type)
            align = trace.shift_shear_gamma()
            align = list(map(float, [align[0,0], align[0,1], align[1,0], align[1,1]]))
            sdata["align"] = align
            sdata["shiftTwist"] = trace.shift_twist_gamma()
            sdata["shiftAxial"] = trace.shift_twist_axial()

        if self.type == "AxialFiber":
            # and ndm >= 3:
            sdata["GJ"] = GJ

        type = "Fiber" # self.type
        model.section(type, tag, **sdata)

        for group in self._material:
            mtag = self._material[group]._mtag
            if mtag is None:
                raise ValueError("Material must be added to model before section.")
            for fiber in shape.create_fibers(warp_type=None, group=group):
                model.fiber(y=fiber["y"], 
                            z=fiber["z"],
                            area=fiber["area"],
                            material=mtag, 
                            section=tag)


class _MultiaxialSection(_Section):
    def __init__(self, type, shape, material, **kwds):
        if not isinstance(material, dict):
            material = {
                None: material
            }
        super().__init__(type, shape, material, **kwds)

        self._shape_model = shape.create_model(materials=material)


    def _add_to_model(self, model, tag):
        shape = self._shape

        mixed_type = self._mixed_type(model)

        sdata = {
            "mixed_type": mixed_type,
        }
        if mixed_type in {"UE", "UG"}:
            trace = self.sv.create_trace(form=mixed_type)
            align = trace.shift_shear_gamma()
            align = list(map(float, [align[0,0], align[0,1], align[1,0], align[1,1]]))
            sdata["align"] = align
            sdata["shiftTwist"] = trace.shift_twist_gamma()
            sdata["shiftAxial"] = trace.shift_twist_axial()


        model.section(self.type, tag, **sdata)

        warp_shapes = self._warp_shapes(model, mixed_type)
        for group in self._material:
            mtag = self._material[group]._mtag
            if mtag is None:
                raise ValueError("Material must be added to model before section.")

            # for fiber in shape.create_fibers(warp_type=self._warp, group=group):
            for fiber in shape._create_fibers(self._shape_model, warp_shapes, group=group):
                model.fiber(**fiber, material=mtag, section=tag)



class _ElasticSection(_Section):
    def __init__(self, type, shape, material, warp=None, **kwds):
        super().__init__("Elastic", shape, material, warp)

        if shape._is_composite:
            self.G = 1 
            self.E = 1
        else:
            self.E = self._material[None]["E"]
            self.G = self._material[None]["G"]

    def _add_to_model(self, model, tag):
        shape = self._shape

        # sdata = {
        #     "mixed_type": self._warp,
        # }

        # if self._warp in {"UE", "UG"}:
        #     trace = shape.create_trace(form=self._warp)
        #     align = trace.shift_shear_gamma()
        #     align = list(map(float, [align[0,0], align[0,1], align[1,0], align[1,1]]))
            # sdata["align"] = align
            # sdata["shiftTwist"] = trace.shift_twist_gamma()
            # sdata["shiftAxial"] = trace.shift_twist_axial()

        cmm = shape.cmm()
        cnn = shape.cnn()
        cnv = shape.cnv()
        cnm = shape.cnm()
        cmw = shape.cmw()
        cvv = shape.cvv()
        cww = shape.cww()
        A = self.EA/self.E
        model.section("ElasticFrame", 
                        tag,
                        E=self.E,
                        G=self.G,
                        A=A,
                        Ay=1*A,
                        Az=1*A,
                        Qy=cnm[0,1],
                        Qz=cnm[2,0],
                        Iy=cmm[1,1]/self.E,
                        Iz=cmm[2,2]/self.E,
                        J = self.GJ/self.G,
                        Cw= cww[0,0],
                        Ry= cnv[1,0],
                        Rz= cnv[2,0],
                        # Sy= cvv[1,1],
                        # Sz= cvv[2,2]
                        Sy= cmw[1,0],#*swch,
                        Sz= cmw[2,0],#*swch
        )



#
#  IKS 
#


#
#  Length 
#
mm           = millimeter   = 0.039370079         # Milimeter
cm           = centimeter   = 0.393700787         # Centimeter
m            = meter        = 39.37007874         # Meter
km           = kilometer    = 39370.07874         # Kilometer
inch                        = 1.0
ft           = foot         = 12.0                # International foot
yd           = yard         = 36.0                # International yard
mi           = mile         = 63360.0             # Mile

#
#  Angle 
#
rad          = radian       = 1.0                 # Radian
arcsec                      = 4.84814e-06         # Arcsecond
arcmin                      = 0.000290888         # Arcminute
deg          = degree       = 0.017453293         # Degrees

#
#  Constants 
#
pi           = pi           = 3.141592654         # Pi

#
#  Time 
#
sec          = second       = 1.0                 # second
min          = minute       = 60.0                # minute
hr           = hour         = 3600.0              # hour
day                         = 86400.0             # day
wk           = week         = 604800.0            # week
yr           = year         = 31536000.0          # year

#
#  Force 
#
N            = newton       = 0.000224809         # Newton (force)
dyn          = dyne         = 2.24809e-09         # Dyne
pdl          = poundal      = 3.10809e-05         # Poundal
lbf          = poundf       = 0.001
kip          = klbf         = 1.0

#
#  Mass 
#
slug                        = 8.33333e-05
lbm          = lbm          = 2.59008e-06         # International avoirdupois pound
gm           = gram         = 5.71015e-09
kg           = kilogram     = 5.71015e-06         # Kilogram
tonne                       = 0.005710147         # Metric tonne
oz           = ounce        = 1.6188e-07          # International avoirdupois ounce

#
#  Pressure 
#
Pa           = pascal       = 1.45038e-07         # Pascal
torr                        = 1.93368e-05
kPa          = kilopascal   = 0.000145038
bar                         = 0.014503774
atm          = atmosphere   = 0.014695949         # Standard atmosphere
MPa          = megapascal   = 0.145037736
GPa          = gigapascal   = 145.0377364
psi                         = 0.001               # Pound-square-inch
ksi                         = 1.0

#
#  Density
#
pcf                         = lbm / ft**3         # Pounds per cubic foot

#
#  Amount 
#
mol          = mole         = 1.0                 # Mole

#
#  Frequency 
#
Hz           = hertz        = 1.0                 

#
#  Area 
#
mm2                         = 0.001550003         # Square millimeter
m2                          = 0.15500031          # Square meter
cm2                         = 1550.0031           # Square centimeter
km2                         = 1550003100.0        # Square kilometer
in2          = inch2        = 1.0                 # Square inch
ft2          = feet2        = 144.0               # Square foot
yd2          = yard2        = 1296.0              # Square yard
mi2          = mile2        = 4014489600.0        # Square mile

#
#  Volume 
#
mm3                         = 6.10237e-05         
m3                          = 0.061023744         
cm3                         = 61023.74409         
km3                         = 61023700000000.0    
in3          = inch3        = 1.0                 
ft3          = foot3        = 1728.0              
cyd          = yard3        = 46656.0             
mi3          = mile3        = 254358000000000.0   

#
#  Velocity 
#
mmps                        = 0.039370079         # Millimeter per second
cps          = cmps         = 0.393700787         # Centimeter per second
mps                         = 39.37007874         # Meter per second
kps                         = 39370.07874         # Kilometer per second
ips          = inchps       = 1.0                 # Inch per second
fps          = footps       = 12.0                # Foot per second
yps                         = 36.0                # Yard per second
mph                         = 17.6                # mile per hour

#
#  Acceleration 
#
mmps2                       = 0.039370079
cps2         = cmps2        = 0.393700787
mps2                        = 39.37007874
kps2                        = 39370.07874
ips2         = inchps2      = 1.0
fps2         = footps2      = 12.0
yps2                        = 36.0
mph2                        = 0.004888889
gravity                     = 386.0885827         # Standard gravity

#
#  Angular Velocity 
#
rpm          = revpm        = 6.283185307         # Revolution per minute
radps                       = 1.0                 # Radian per second



#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//


def check_mass(model):
    """
    Compute the total mass of the model.

    This is done by defining a 
    """

def _buckle_factor(boundary, phi=0):
    import math
    if boundary == "pin-pin":
        return math.pi

    if boundary == "fix-slide":
        return math.pi

    if boundary == "fix-fix":
        return 2*math.pi

    if boundary == "fix-pin":
        import scipy.optimize
        f = lambda x: math.tan(x) - x/(1 + x**2*phi/12)
        sol = scipy.optimize.root_scalar(f, x0=0.7, bracket=(math.pi, 1.45*math.pi))
        if sol.converged:
            return sol.root

    if boundary == "fix-free":
        return math.pi/2

    if boundary == "pin-slide":
        return math.pi/2

def _fix_node(model, node, type):
    ndf = model.getNDF()
    reactions = [0 for _ in range(ndf)]

    long, tran, vert = range(3)
    if ndf == 6:
        bend = vert+3
        # always fix out-of-plane rotation, which spins
        # about the transverse DOF
        reactions[tran+3] = 1
        reactions[long+3] = 1 # torsion
        reactions[2] = 1
    else:
        bend = vert


    if node > 1:
        vert = 0
    else:
        vert = 1

    if type == "fix":
        reactions[tran] = 1
        reactions[long] = 0 if node > 1 else 1
        reactions[bend] = 1

    elif type == "pin":
        reactions[tran] = 1
        reactions[long] = 0 if node > 1 else 1
        reactions[bend] = 0

    elif type == "slide":
        reactions[tran] = 0
        reactions[long] = 0 if node > 1 else 1
        reactions[bend] = 1

    elif type == "free":
        pass

    model.fix(node, *reactions)
    return reactions


def create_prism(length:    float,
                 element:   str,
                 section:   dict,
                 boundary:  tuple,
                 geometry:  str = None,
                 transform: str = None,
                 divisions: int = 1,
                 rotation = None, 
                 ndm=3):
    import opensees.openseespy as ops

    L  = length

    # Number of elements discretizing the column
    ne = divisions

    elem_type  = element
    geom_type  = transform

    # Number of integration points along each element
    nIP = 5
    nn = ne + 1

    model = ops.Model(ndm=ndm)

    for i in range(1, nn+1):
        x = (i-1)/float(ne)*L
        if ndm == 3:
            location = (x, 0.0, 0.0)
        else:
            location = (x, 0.0)

        if rotation is not None:
            location = tuple(rotation@location)

        model.node(i, location)

        model.mass(i, *[1.0]*model.getNDF())


    # Define boundary conditions
    if isinstance(boundary[0], str):
        _fix_node(model,  1, boundary[0])
    else:
        model.fix(1, boundary[0])

    if isinstance(boundary[1], str):
        _fix_node(model, nn, boundary[1])
    else:
        model.fix(nn, boundary[1])

    #
    # Define cross-section 
    #
    sec_tag = 1
    properties = []
    for k,v in section.items():
        properties.append("-" + k)
        properties.append(v)

    model.section("FrameElastic", sec_tag, *properties)

    # Define geometric transformation
    geo_tag = 1
    if ndm == 3:
        vector = (0,  0, 1)
#           vector = (0,  -1, 0)
        if rotation is not None:
            vector = tuple(map(float, rotation@vector))
        # for |_ 
    else:
        vector = ()

    model.geomTransf(geom_type, geo_tag, *vector)

    # Define elements
    for i in range(1, ne+1):
        if geometry is None or geometry == "Linear" or "Exact" in elem_type:
            model.element(elem_type, i, (i, i+1),
                        section=sec_tag,
#                       shear=1,
                        transform=geo_tag)
        else:
            model.element(elem_type, i, (i, i+1),
                        section=sec_tag,
                        order={"Linear": 0, "delta": 1}[geometry],
                        transform=geo_tag)

    return model

class Prism:
    def __init__(self,
                 length:    float,
                 element:   str,
                 section:   dict,
                 boundary:  tuple,
                 geometry:  str = None,
                 transform: str = None,
                 divisions: int = 1,
                 rotation = None,
                 shear = True):

        self.length    = length
        self.element   = element
        self.section   = section
        self.boundary  = boundary
        self.geometry  = geometry
        self.transform = transform
        self.divisions = divisions
        self.use_shear = shear
        self.rotation = rotation

    def create_model(self, ndm=3, file=None):
        import opensees.openseespy as ops

        L  = self.length
        boundary = self.boundary

        # Number of elements discretizing the column
        ne = self.divisions

        elem_type  = self.element
        geom_type  = self.transform

        # Number of integration points along each element
        nIP = 5
        nn = ne + 1
        if file is not None:
            file = open(file, "w+")
        model = ops.Model(ndm=ndm, echo_file=file)

        for i in range(1, nn+1):
            x = (i-1)/float(ne)*L
            if ndm == 3:
                location = (x, 0.0, 0.0)
            else:
                location = (x, 0.0)

            if self.rotation is not None:
                location = tuple(self.rotation@location)

            model.node(i, location)

            model.mass(i, *[1.0]*model.getNDF())


        # Define boundary conditions
        if isinstance(boundary[0], str):
            _fix_node(model,  1, boundary[0])
        else:
            model.fix(1, boundary[0])

        if isinstance(boundary[1], str):
            _fix_node(model, nn, boundary[1])
        else:
            model.fix(nn, boundary[1])

        #
        # Define cross-section 
        #
        sec_tag = 1
        properties = []
        for k,v in self.section.items():
            properties.append("-" + k)
            properties.append(v)

        model.section("ElasticFrame", sec_tag, *properties)

        # Define geometric transformation
        geo_tag = 1
        if ndm == 3:
            vector = (0,  0, 1)
#           vector = (0,  -1, 0)
            if self.rotation is not None:
                vector = tuple(map(float, self.rotation@vector))
            # for |_ 
        else:
            vector = ()

        model.geomTransf(geom_type, geo_tag, *vector)

        # Define elements
        for i in range(1, ne+1):
            if self.geometry is None or self.geometry == "Linear" or "Exact" in elem_type:
                model.element(elem_type, i, (i, i+1),
                            section=sec_tag,
                            transform=geo_tag,
#                           iter=(2, 1e-12),
                            shear=int(self.use_shear))
            else:
                model.element(elem_type, i, (i, i+1),
                            section=sec_tag,
                            order={"Linear": 0, "delta": 1}[self.geometry],
                            transform=geo_tag,
                            shear=int(self.use_shear))

        return model

    def buckle(self, model):
        # Define loads
        if self.use_shear:
            phi = 12*E*I/(Ay*G*L**2)
            lam = _buckle_factor(boundary, phi)
            kL = L/lam
            euler_load = E*I/kL**2  / (1 + lam**2*phi/12)
        else:
            kL = L/_buckle_factor(boundary)
            euler_load = E*I/kL**2

        model.pattern("Plain", 1, "Linear")
        if self.ndm == 2:
            load = (-euler_load, 0.0, 0.0)
        else:
            load = (-euler_load, 0.0, 0.0, 0, 0, 0)

        model.load(nn, *load, pattern=1)

        return model


def node_average(model, response, ndm=2, keys=None):
    """
    Return a dictionary of stress components per node.
    For 2D: sxx, syy, sxy
    For 3D: sxx, syy, szz, sxy, syz, sxz
    Summed values from each element's node contribution are averaged at the end.

    Args:
        model: an OpenSees model-like object with the following interface:
               - getEleTags()
               - getEleClassTags(eleTag)
               - getNodeTags()
               - eleNodes(eleTag)
               - eleResponse(eleTag, str)
        response (str): a label if needed for e.g. 'stress', 'stressAtNodes'
        ndm (int): number of dimensions (2 or 3)

    Returns:
        dict[nodeTag] = {
            "xx": float, 
            "yy": float,
            (optionally in 2D) "xy": float,
            (optionally in 3D) "zz", "yz", "xz",
            ... possibly derived quantities ...
        }
    """

    # Helper: pick which elements are 2D vs 3D, if needed
    # For example, you might have separate function(s) or logic:
    #    ele_tags = _stress_2d_ele_tags_only(model, model.getEleTags())
    # or for 3D:
    #    ele_tags = _stress_3d_ele_tags_only(model, model.getEleTags())
    #
    # Below, we simply re-use the user's helper for 2D, but you'll have to define
    # your own for 3D or unify them if your model can contain both:
    ele_tags_all = model.getEleTags()
    if ndm == 2:
        ele_tags = _stress_2d_ele_tags_only(model, ele_tags_all)
    else:
        # Placeholder; you must implement a 3D filter or skip if you already know
        # the elements are 3D:
        ele_tags = ele_tags_all

    # For a typical 2D element with 3 stress components: sxx, syy, sxy
    # For a typical 3D element with 6 stress components: sxx, syy, szz, sxy, syz, sxz
    # You may need to adapt this depending on your element type(s).
    if ndm == 2:
        keys = "sxx", "syy", "sxy"
    else:
        keys = "sxx", "syy", "szz", "sxy", "syz", "sxz"


    node_tags = model.getNodeTags()

    # Initialize a dictionary to hold stress sums and a count of how many
    # times each node is encountered (to do averaging).
    sig_dict = {node: {"_count": 0, **{key: 0.0 for key in keys}} for node in node_tags}

    nrc = len(keys)

    # We assume each element’s node-level stress output is available:
    # e.g. model.eleResponse(elem_tag, 'stressAtNodes')
    # and yields a flat array (with length = n_nodes_in_element * nrc).
    # Then we reshape it to (nen, nrc).
    for elem in ele_tags:
        ele_node_tags = model.eleNodes(elem)
        nen = len(ele_node_tags)

        sig_per_node = np.reshape(model.eleResponse(elem, response), (nen, nrc))

        # Accumulate into sig_dict
        for i, node in enumerate(ele_node_tags):
            for j, key in enumerate(keys):
                sig_dict[node][key] += sig_per_node[i][j]

            sig_dict[node]["_count"] += 1

    # Now do averaging (if a node belongs to multiple elements).
    for node in node_tags:
        c = sig_dict[node]["_count"] or 1.0

        for key in keys:
            sig_dict[node][key] /= c


    # Clean up
    for node in node_tags:
        sig_dict[node].pop("_count", None)

    return sig_dict


def find_node(model, x=None, y=None, z=None, tol=1e-8):
    """
    Return the tag of the *first* node that matches the specified coordinates
    within a given tolerance. If no match is found, returns None.

    Parameters
    ----------
    model : opensees.openseespy.Model
        The model object containing nodes.
    x : float, optional
        X-coordinate to match (if None, no X match is required).
    y : float, optional
        Y-coordinate to match (if None, no Y match is required).
    tol : float
        Matching tolerance for coordinates.

    Returns
    -------
    int or None
        The tag of the first matching node, or None if not found.
    """
    # Get the list of all node tags in the model
    node_tags = model.getNodeTags()

    for tag in node_tags:
        coords = model.nodeCoord(tag)  # [xCoord, yCoord, ...]
        # Check requested coordinates
        if x is not None and abs(coords[0] - x) > tol:
            continue

        if y is not None and abs(coords[1] - y) > tol:
            continue

        if z is not None and abs(coords[2] - z) > tol:
            continue

        # If we've passed all checks, this node matches
        return tag

    # If no node passes the test, return None
    return None


def find_nodes(model, x=None, y=None, z=None, tol=1e-8):
    """
    Return an iterator of node tags that match the specified coordinates
    within a given tolerance.

    Parameters
    ----------
    model : opensees.openseespy.Model
        The model object containing nodes.
    x : float, optional
        X-coordinate to match (if None, no X match is required).
    y : float, optional
        Y-coordinate to match (if None, no Y match is required).
    tol : float
        Matching tolerance for coordinates.

    Yields
    ------
    int
        The tag of each matching node.
    """
    node_tags = model.getNodeTags()
    for tag in node_tags:
        coords = model.nodeCoord(tag)  # [xCoord, yCoord, ...]
        # Check X coordinate if requested
        if x is not None and abs(coords[0] - x) > tol:
            continue
        # Check Y coordinate if requested
        if y is not None and abs(coords[1] - y) > tol:
            continue
        # Check Z coordinate if requested
        if z is not None and abs(coords[2] - z) > tol:
            continue
        # If we've passed all checks, this node matches
        yield tag


# class Analysis:
#     def __init__(self, model):
#         self._model = model
#         self._test = None

#         self._integrator = None

#     def integrator(self, *args):
#         self._integrator = args

#     def increment(self, factor=1.0):
#         self._model.integrator("LoadControl", factor)
#         self._model.analysis("Static")
#         self._model.analyze(1, operation='increment')

#     def iteration(self):
# #       self._model.test("FixedNumIter", 1, 1)
#         niter = 1
#         tol = 1e-12
#         self._model.test("EnergyIncr", tol, niter, 5)
#         self._model.analysis("Static")
#         self._model.analyze(1, operation='iteration')




#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//

class Pattern:
    _persistent: bool

    def _activate(self, model):
        pass

    def _deactivate(self, model):
        tag = self.tag
        if self._persistent:
            model.loadConst(tag=tag)
        else:
            model.removePattern(tag=tag)


class AmbientAcceleration(Pattern):
    def __init__(self, dof, series, dt=None, time=None, persist=False, tag=None):
        self.tag = tag
        self._dof = dof
        if hasattr(series, "tolist"):
            series = series.tolist()
        self._data = series
        self._dt = dt
        # super().__init__(self, persist=persist)


    def _activate(self, model, pattern=None):
        if pattern is None:
            pattern = 1

        series_tag = 1 + 6*(pattern - 1)

        if pattern not in model._patterns:
            model.timeSeries("Path", series_tag,
                             dt=self._dt, values=self._data)
            model.pattern("UniformExcitation",
                          pattern, self._dof, accel=series_tag)
            series_tag += 1
            # model._patterns.append(pattern)
        else:
            model.activatePattern(tag=pattern)






#
#  MKS 
#

#  https://en.wikipedia.org/wiki/MKS_units



#
#  Length 
#
mm           = millimeter   = 0.001               # Milimeter
cm           = centimeter   = 0.01                # Centimeter
m            = meter        = 1.0                 # Meter
km           = kilometer    = 1000.0              # Kilometer
inch                        = 0.0254              
ft           = foot         = 0.3048              # International foot
yd           = yard         = 0.9144              # International yard
mi           = mile         = 1609.344            # Mile

#
#  Angle 
#
rad          = radian       = 1.0                 # Radian
arcsec                      = 4.84814e-06         # Arcsecond
arcmin                      = 0.000290888         # Arcminute
deg          = degree       = 0.017453293         # Degrees

#
#  Constants 
#
pi           = pi           = 3.141592654         # Pi

#
#  Time 
#
sec          = second       = 1.0                 # second
min          = minute       = 60.0                # minute
hr           = hour         = 3600.0              # hour
day                         = 86400.0             # day
wk           = week         = 604800.0            # week
yr           = year         = 31536000.0          # year

#
#  Force 
#
N            = newton       = 1.0                 # Newton (force)
dyn          = dyne         = 1e-05               # Dyne
pdl          = poundal      = 0.138254954         # Poundal
lbf          = poundf       = 4.448221657         
kip          = klbf         = 4448.221657         

#
#  Mass 
#
slug                        = 14.59390307
lbm          = lbm          = 0.453592368         # International avoirdupois pound
gm           = gram         = 0.001
kg           = kilogram     = 1.0                 # Kilogram
tonne                       = 1000.0              # Metric tonne
oz           = ounce        = 0.028349523         # International avoirdupois ounce

#
#  Pressure 
#
Pa           = pascal       = 1.0                 # Pascal
torr                        = 133.3223684         
kPa          = kilopascal   = 1000.0              
bar                         = 100000.0            
atm          = atmosphere   = 101325.0            # Standard atmosphere
MPa          = megapascal   = 1000000.0           
GPa          = gigapascal   = 1000000000.0        
psi                         = 6894.757358         # Pound-square-inch
ksi                         = 6894757.358         

#
#  Density 
#
pcf                         = lbm / ft**3         # Pounds per cubic foot

#
#  Amount 
#
mol          = mole         = 1.0                 # Mole

#
#  Frequency 
#
Hz           = hertz        = 1.0                 

#
#  Area 
#
mm2                         = 1e-06               # Square millimeter
m2                          = 0.0001              # Square meter
cm2                         = 1.0                 # Square centimeter
km2                         = 1000000.0           # Square kilometer
in2          = inch2        = 0.00064516          # Square inch
ft2          = feet2        = 0.09290304          # Square foot
yd2          = yard2        = 0.83612736          # Square yard
mi2          = mile2        = 2589988.11          # Square mile

#
#  Volume 
#
mm3                         = 1e-09               
m3                          = 1e-06               
cm3                         = 1.0                 
km3                         = 1000000000.0        
in3          = inch3        = 1.63871e-05         
ft3          = foot3        = 0.028316847         
cyd          = yard3        = 0.764554858         
mi3          = mile3        = 4168181825.0        

#
#  Velocity 
#
mmps                        = 0.001               # Millimeter per second
cps          = cmps         = 0.01                # Centimeter per second
mps                         = 1.0                 # Meter per second
kps                         = 1000.0              # Kilometer per second
ips          = inchps       = 0.0254              # Inch per second
fps          = footps       = 0.3048              # Foot per second
yps                         = 0.9144              # Yard per second
mph                         = 0.44704             # mile per hour

#
#  Acceleration 
#
mmps2                       = 0.001               
cps2         = cmps2        = 0.01                
mps2                        = 1.0                 
kps2                        = 1000.0              
ips2         = inchps2      = 0.0254              
fps2         = footps2      = 0.3048              
yps2                        = 0.9144              
mph2                        = 0.000124178         
gravity                     = 9.80665             # Standard gravity

#
#  Angular Velocity 
#
rpm          = revpm        = 6.283185307         # Revolution per minute
radps                       = 1.0                 # Radian per second




def check(a, b):
    
    print(f"{a-b}")


def test(u):

    # check(u.oz, u.lbm/16)

    check(u.lbf/u.slug,   u.ft / u.sec**2)
    check(u.pdl,   0.138254954376*u.N)
    check(u.slug, 32.17405*u.lbm)
    check(u.slug,  14.59390*u.kg)
    check(u.lbm, 0.45359237*u.kg)  # International avoirdupois pound




if __name__ == "__main__":
    from xara.units import si, iks, ips, fks, fps

    test(si)
    test(iks)
    test(ips)
    test(fks)
    test(fps)


import xara
from xara import Section
from dataclasses import dataclass


def create_section(model, type, tag, shape, material, warp=True, shear=True, center=None):

    if type == "Elastic":
        cmm = shape.cmm()
        cnn = shape.cnn()
        cnv = shape.cnv()
        cnm = shape.cnm()
        cmw = shape.cmw()
        cww = shape.cww()
        cvv = shape.cvv()
        A = cnn[0,0]
        model.section("ElasticFrame", tag,
                        E=material["E"],
                        G=material["G"],
                        A=cnn[0,0],
                        # Ay=1*A,
                        # Az=1*A,
                        # Qy=cnm[0,1],
                        # Qz=cnm[2,0],
                        Iy=cmm[1,1],
                        Iz=cmm[2,2],
                        J =shape._analysis.torsion_constant(),
                        Cw= cww[0,0],
                        Ry= cnv[1,0],
                        Rz= cnv[2,0],
                        Sy= cvv[1,1],
                        Sz= cvv[2,2]
        )

    elif type == "Uniaxial":
        pass

    else:
        if isinstance(material,int):
            mtag = material
        else:
            mtag = 1
            if material is None:
                material = {"name": "ElasticIsotropic", "E": 1, "G": 1}
            if "name" in material:
                mtype = material["name"]
                del material["name"]
            else:
                mtype = "ElasticIsotropic"

            model.material(mtype, mtag, **material)

        model.section("ShearFiber", tag, GJ=0)

        for fiber in shape.create_fibers(warp=warp, shear=shear):
            model.fiber(**fiber, material=mtag, section=tag)



@dataclass
class Prism:
    length: float
    shape: str
    material: dict
    boundary: tuple
    element: str = "PrismFrame"
    section: str = "Elastic"
    transform: str = "Linear"
    geometry : str = None
    vertical: int = 3
    rotation: tuple = None
    integration: dict = None
    divisions: int = 1
    order: int = 1
    shear: int = 1
    shear_warp: bool = False
    warp: list = None
    mass: float = None
    joint_offset: dict=None


    def __repr__(self):
        return f"""\
        Prism(length={self.length}, 
              shape={self.shape},
              material={self.material}, 
              boundary={self.boundary}, 
              element={self.element}, 
              section={self.section}, 
              transform={self.transform}, 
              geometry={self.geometry}, 
              vertical={self.vertical}, 
              rotation={self.rotation}, 
              divisions={self.divisions}, 
              order={self.order}, 
              shear={self.shear}
              warp={self.warp})
        """


    def create_model(self,
                    geometry:  str = None,
                    echo_file=None,
                    **options) -> xara.Model:

        L  = self.length
        element = self.element
        shape = self.shape
        material = self.material
        boundary = self.boundary
        transform = self.transform
        vertical = self.vertical
        divisions = self.divisions
        rotation = self.rotation
        section = self.section
        order = self.order
        shear = self.shear
        geometry = self.geometry
        warp = self.warp
        if warp is None:
            warp = [0,0,0]




        # Number of elements discretizing the column
        ne = divisions

        nen = order + 1
        nn = ne*(nen-1)+1

        model = xara.Model(ndm=3, ndf=6+sum(warp), echo_file=echo_file)

        for i in range(1, nn+1):
            x = (i-1)*L/float(nn-1)
            location = (x, 0, 0)

            if rotation is not None:
                location = tuple(rotation@location)

            model.node(i, location)


        # Define boundary conditions
        model.fix( 1, boundary[0])
        model.fix(nn, boundary[1])

        #
        # Define cross-section 
        #
        if isinstance(shape, dict):
            model.section("ElasticFrame", 1, **shape, **material)
        elif hasattr(section, "_add_to_model"):
            if isinstance(material, dict):
                for i, mat in enumerate(material.values(), start=1):
                    model.material(mat, i)
            else:
                model.material(material, 1)
            model.section(section, 1)
        elif callable(section):
            section(model, 1, shape, material)
        else:
            shape.send_xara(model,
                            section, 1,
                            material,
                            warp=warp or self.shear_warp,
                            shear=self.shear_warp)

        # Define geometric transformation
        if vertical == 3:
            orient = (0, 0, 1)
        if rotation is not None:
            orient = tuple(map(float, rotation@orient))

        if self.joint_offset is not None:
            model.geomTransf(transform, 1, orient, offset=self.joint_offset)
        else:
            model.geomTransf(transform, 1, *orient)


        # Define elements
        # options = {}
        if self.mass is not None:
            options["mass"] = self.mass

        for i in range(ne):
            start = i * (nen - 1) + 1
            nodes = list(range(start, start + nen))
            if "Exact" in element:
                model.element(element, i+1, 
                            nodes,
                            section=1,
                            shear=shear,
                            transform=1, **options)
            elif self.integration is not None:
                model.element(element, i+1, nodes,
                            section=1,
                            shear=shear,
                            order={"Linear": 0, "delta": 1}[geometry or "Linear"],
                            gauss_type=f"{self.integration['type']}",
                            gauss_points=self.integration['points'],
                            transform=1, 
                            **options)
            else:
                model.element(element, i+1, nodes,
                            section=1,
                            shear=shear,
                            order={"Linear": 0, "delta": 1}[geometry or "Linear"],
                            transform=1, 
                            **options)

        return model





import sys
from pathlib import Path
from itertools import chain

_PREAMBLE = r"""

"""


class Evaluation:
    def __init__(self, paths, exclusions=None):
        self._exclusions = exclusions if exclusions is not None else []
        self._paths = paths


    def run(self):
        for path in self._paths:
            if not path.is_file():
                continue
            print(path)
            try:
                TestCase(path, self).run()

            except Exception as e:
                print(f"Error running test case {path}: {e}")
                raise e
                continue


class TestCase:
    def __init__(self, file, evaluation=None):
        self._file = file
        self._evaluation = evaluation


    def run(self):
        from opensees import tcl
        with open(self._file, "r") as f:
            script = _PREAMBLE + f.read()

        interp = tcl.Interpreter()

        def verify(*args):
            # interp.eval("puts {Verifying: " + " ".join(str(arg) for arg in args) + "}")
            objects = {
                "element": set(),
                "section": set(),
                "material": set()
            }

            if False: #args[1] == "value":
                try:
                    model = interp.serialize()["StructuralAnalysisModel"]
                    for element in model["geometry"].get("elements", []):
                        objects["element"].add(element["type"])

                    for section in model["properties"].get("SectionForceDeformation", []):
                        objects["section"].add(section["type"])

                except Exception as e:
                    print(f"Error serializing model: {e}")
                    pass


            interp.eval("OpenSees::verify " + " ".join(f'"{arg}"' for arg in args))

        interp._tcl.createcommand("verify", verify)

        interp.eval(script)
        interp.eval("wipe")
        interp.eval("_wipeNodes")
        return {"status": "success"}

def run(args):

    if len(args) == 1:
        files = chain(
                  Path(".").rglob("test*.tcl"),
                  Path(".").rglob("o-*.tcl"),
                  Path(".").rglob("x-*.tcl")
                )
    else:
        files = (Path(arg) for arg in args[1:])

    Evaluation(files).run()
    return



if __name__ == "__main__":

    run(sys.argv)



#===----------------------------------------------------------------------===//
#
#                                   xara
#                              https://xara.so
#
#===----------------------------------------------------------------------===//
#
# Copyright (c) 2025, OpenSees/Xara Developers
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#===----------------------------------------------------------------------===//
from dataclasses import dataclass



class Pattern:   pass # Acceleration, Element, Boundary, Constraint

class Sequence:  pass


class Search:    pass
class Algorithm: pass
class Test:      pass

class ConstraintHandler: pass


class EigenSolver: pass
class LinearSolver: pass


_EquilibriumResponse = [
    "residual",
    "tangent",

    "position",
    "rotation",
    "velocity",
    "acceleration",

    "resistance", # "force",
    "reaction"
]

_SpectralResponse = [
    "frequency",
    "period",
    "mode"
]

class Analysis:
    Success = 0
    Failure = 1
    Partial = 2 # in progress


    _status: int
    _constraints: ConstraintHandler

    def _analysis_cached(self, model):
        
        if not hasattr(model, "_last_analysis") or \
            model._last_analysis is not self:
            model._last_analysis = self 
            return False 
        
        return True


    def solution(self,
                 response: str, 
                 node=None, dof=None, 
                 element=None, material=None, section=None,
                 **kwds):
        
        if node is None and element is None:
            return self._solution_vector()
        
        if element is not None and node is not None:
            raise ValueError("Cannot specify both element and node for solution retrieval.")

        if element is not None:
            return self._element_solution(
                response, element=element, material=material, section=section
            )
        
        if node is not None:
            return self._node_solution(
                response, node=node, dof=dof
            )

    def gradient(self, response, parameter, **loc):
        pass

    def jacobian(self, response, parameters):
        pass


class StabilityAnalysis(Analysis):
    def __init__(self,
                 constraints=None,
                 solver: "EigenSolver" = None):
        self._status = Analysis.Partial
        self._constraints = constraints


class EigenAnalysis(Analysis):
    """ 
    Perform an eigenvalue analysis to determine the natural frequencies and mode shapes of the structure.
    """
    def __init__(self, n, highest, unorm:float):
        self._status = Analysis.Partial
        self._n = n
        self._highest = highest
        self._unorm = unorm

    def analyze(self, model):
        model.modalProperties()

        self._status = Analysis.Success

        return self._status



class SpectrumAnalysis(Analysis):
    def __init__(self,
                 sequence, 
                 modes: EigenAnalysis,
                 highest=None, unorm:float=None,
                 solver: EigenSolver=None):

        self._status = Analysis.Partial
        self._constraints = None
        self._modes = modes
        self._sequence = sequence
        self._solver = solver
        self._highest = highest
        self._unorm = unorm




class _EquilibriumAnalysis(Analysis):
    _algorithm : Algorithm
    _solver    : LinearSolver
    _test      : Test
    _search    : Search
    _response = [
        *_EquilibriumResponse
    ]

    def __init__(self):
        self._patterns = []
        self._time = 0.0

    def _add_pattern(self, pattern, type=None):
        self._patterns.append((pattern, type))

    def _start_equilibrium(self,model):
        if not self._analysis_cached(model):
            model.wipeAnalysis()
            if self._constraints is not None:
                model.constraints(self._constraints)
            if self._test is not None:
                model.test(*self._test)
            if self._solver is not None:
                model.solver(self._solver)
                

        # Patterns
        model.setTime(self._time)
        for i,(pattern, type) in enumerate(self._patterns):
            pattern._activate(model, pattern=i+1)

        # Algorithm
        if self._algorithm is None:
            model.algorithm("Newton")
        else:
            model.algorithm(*self._algorithm)



class StaticAnalysis(_EquilibriumAnalysis):
    """
    A static analysis solves for nodal displacements under static loading conditions.
    Parameters
    ----------
    model : :py:class:`Model`
        The finite element model to be analyzed.
    pattern : Pattern
        The load pattern to be applied during the analysis.
    step : float, optional
        The load step size, by default 1.0.
    constraints : ConstraintHandler, optional
        The constraint handler to manage boundary conditions, by default None.
    test : Test, optional
        The convergence test to determine when the solution has converged.
    algorithm : Algorithm, optional
        The solution algorithm to be used, by default Newton-Raphson.
    """
    def __init__(self,
                 pattern: Pattern = None,
                 step:    float=1.0,
                 # Analysis
                 constraints=None,
                 # EquilibriumAnalysis
                 test: Test = None,
                 algorithm: Algorithm=None,
                 search=None,
                 solver:    LinearSolver=None,
                 ):

        self._status = Analysis.Partial
        self._constraints = constraints

        self._algorithm = algorithm
        self._solver    = solver
        self._test      = test

        super().__init__()

        if pattern is None:
            pass
        elif isinstance(pattern, list):
            for p in pattern:
                self._add_pattern(p, "Linear")
        else:
            self._add_pattern(pattern, "Linear")


    def analyze(self, model):
        status = Analysis.Partial
        # add our patterns
        self._start_equilibrium(model)

        # analyze
        model.analysis("Static")
        status = model.analyze(1)

        # clean up
        if status == 0:
            self._status = Analysis.Success
        else:
            self._status = Analysis.Failure
        model.wipeAnalysis()
        model.loadConst(time=self._time)
        return status



class DynamicAnalysis(_EquilibriumAnalysis):
    """DynamicAnalysis(model, pattern, **options)

    Perform a dynamic analysis to determine the time-dependent response of the structure.

    Parameters
    ----------
    model : Model
        The finite element model to be analyzed.
    pattern : Pattern
        The load patterns to be advanced during the analysis.
    step : float
        The time step size for the analysis.
    """
    class Integrator: pass 


    def __init__(self,
                 pattern: Pattern=None,
                 step=None, step_scale=None,
                 integrator: "DynamicAnalysis.Integrator"=None,
                 # Analysis
                 constraints=None,
                 # EquilibriumAnalysis
                 test: Test = None,
                 algorithm: Algorithm=None,
                 search=None,
                 solver:    LinearSolver=None,
                ):

        self._status = Analysis.Partial
        self._constraints = constraints

        self._algorithm = algorithm
        self._solver    = solver
        self._test      = test

        super().__init__()

        if isinstance(pattern, list):
            for p in pattern:
                self._add_pattern(p, None)
        elif pattern is not None:
            self._add_pattern(pattern, None)

        self._integrator = integrator
        # self._add_series(series)


    def analyze(self, model, dt, steps=1):
        # self._start_equilibrium(model)

        # model.constraints('Transformation')

        model.integrator('Newmark', 0.5, 0.25)
        model.analysis('Transient')
        return model.analyze(steps, dt)


class ControlAnalysis(_EquilibriumAnalysis):
    # DispControl, LoadControl, ArcLength
    def __init__(self,
                 pattern: Pattern=None,
                 step : float=None,
                 control=None,
                 min_step: float=None,
                 max_step: float=None,
                 # Analysis
                 constraints=None,
                 # EquilibriumAnalysis
                 test: Test=None,
                 algorithm: Algorithm=None,
                 search=None,
                 solver: LinearSolver=None,
                 #
                 **kwds):

        self._status = Analysis.Partial
        self._constraints = constraints

        self._algorithm = algorithm
        self._solver    = solver
        self._test      = test

        super().__init__()
        self._add_pattern(pattern, "Linear")

        self._control = control
        self._step = step
        self._control_kwds = kwds

    def update(self, **integrator):
        pass

    def analyze(self, model, repeat:int=1)->int:
        self._start_equilibrium(model)
        status = Analysis.Partial
        model.loadConst(time=self._time)

        # add our patterns

        # analyze

        self._time = model.getTime()
        model.loadConst(time=self._time)



def Objective(builder:  callable, # | Model,
              analysis: callable, # | Analysis | Pattern, 
              response):

    # if analysis is a pattern, create a basic static analysis

    def evaluate(self):
        pass

    def gradient(self):
        pass


class AnalysisSequence:
    pass


if __name__ == "__main__":
    import sys
    import xara

    model = xara.Model()


    options = {
        "algorithm": {"type": "NewtonRaphson"},
        "test":      {"type": "Residual",  "tolerance": 1e-6,  "iterate": 10 },
        "solver":    "Umfpack",
        "constraints": "Transformation"
    }

    StaticAnalysis(pattern="SelfWeight", **options).analyze(model)


    ma = EigenAnalysis(model, 3)
    ma.analyze()
    model.damping(ma, {1: 0.05, 2: 0.05, 3: 0.05}) # TODO


    motion = Pattern(persist=False)
    rha = DynamicAnalysis(pattern=motion, step=0.01, **options)



    rha.analyze(model)

    g = Objective(model,  rha,  "u", node=1, dof=1)

    K = rha.jacobian("p", "u")
    M = rha.jacobian("p", "a")
    C = rha.jacobian("p", "v")

    u = rha.solution("u", node=1, dof=1)

    s = rha.solution("s", element=1, material=1)



#
#  CGS 
#


#
#  Length 
#
mm           = millimeter   = 0.1                 # Milimeter
cm           = centimeter   = 1.0                 # Centimeter
m            = meter        = 100.0               # Meter
km           = kilometer    = 100000.0            # Kilometer
inch                        = 2.54                
ft           = foot         = 30.48               # International foot
yd           = yard         = 91.44               # International yard
mi           = mile         = 160934.4            # Mile

#
#  Angle 
#
rad          = radian       = 1.0                 # Radian
arcsec                      = 4.84814e-06         # Arcsecond
arcmin                      = 0.000290888         # Arcminute
deg          = degree       = 0.017453293         # Degrees

#
#  Constants 
#
pi           = pi           = 3.141592654         # Pi

#
#  Time 
#
sec          = second       = 1.0                 # second
min          = minute       = 60.0                # minute
hr           = hour         = 3600.0              # hour
day                         = 86400.0             # day
wk           = week         = 604800.0            # week
yr           = year         = 31536000.0          # year

#
#  Force 
#
N            = newton       = 100000.0            # Newton (force)
dyn          = dyne         = 1.0                 # Dyne
pdl          = poundal      = 13825.49538         # Poundal
lbf          = poundf       = 444822.1657         
kip          = klbf         = 444822165.7         

#
#  Mass 
#
slug                        = 14593.90307         
lbm          = lbm          = 453.592368          # International avoirdupois pound
gm           = gram         = 1.0                 
kg           = kilogram     = 1000.0              # Kilogram
tonne                       = 1000000.0           # Metric tonne
oz           = ounce        = 28.349523           # International avoirdupois ounce

#
#  Pressure 
#
Pa           = pascal       = 10.0                # Pascal
torr                        = 1333.223684         
kPa          = kilopascal   = 10000.0             
bar                         = 1000000.0
atm          = atmosphere   = 1013250.0           # Standard atmosphere
MPa          = megapascal   = 10000000.0          
GPa          = gigapascal   = 10000000000.0       
psi                         = 68947.57358         # Pound(f)-square-inch
ksi                         = 68947573.58         

#
#  Density 
#
pcf                         = lbm / ft**3         # Pounds per cubic foot

#
#  Amount 
#
mol          = mole         = 1.0                 # Mole

#
#  Frequency 
#
Hz           = hertz        = 1.0                 

#
#  Area 
#
mm2                         = 0.01                # Square millimeter
m2                          = 1.0                 # Square meter
cm2                         = 10000.0             # Square centimeter
km2                         = 10000000000.0       # Square kilometer
in2          = inch2        = 6.4516              # Square inch
ft2          = feet2        = 929.0304            # Square foot
yd2          = yard2        = 8361.2736           # Square yard
mi2          = mile2        = 25899881103.0       # Square mile

#
#  Volume 
#
mm3                         = 0.001               
m3                          = 1.0                 
cm3                         = 1000000.0           
km3                         = 1000000000000000.0  
in3          = inch3        = 16.387064           
ft3          = foot3        = 28316.84659         
cyd          = yard3        = 764554.858          
mi3          = mile3        = 4168180000000000.0  

#
#  Velocity 
#
mmps                        = 0.1                 # Millimeter per second
cps          = cmps         = 1.0                 # Centimeter per second
mps                         = 100.0               # Meter per second
kps                         = 100000.0            # Kilometer per second
ips          = inchps       = 2.54                # Inch per second
fps          = footps       = 30.48               # Foot per second
yps                         = 91.44               # Yard per second
mph                         = 44.704              # mile per hour

#
#  Acceleration 
#
mmps2                       = 0.1                 
cps2         = cmps2        = 1.0                 
mps2                        = 100.0               
kps2                        = 100000.0            
ips2         = inchps2      = 2.54                
fps2         = footps2      = 30.48               
yps2                        = 91.44               
mph2                        = 0.012417778         
gravity                     = 980.665             # Standard gravity

#
#  Angular Velocity 
#
rpm          = revpm        = 6.283185307         # Revolution per minute
radps                       = 1.0                 # Radian per second



from typing import overload, Literal


class _Materials:
    def material(self, type: str, tag: int, *args, **kwargs):
        """
        Create a material of the specified type.
        """

    def section(self, type: str, tag: int, *args, **kwargs):
        """
        Create a section of the specified type.
        """

class _Elements:

    @overload
    def element(self, type: str, tag: int, nodes: tuple, **kwargs) -> int:
        """
        Create an element of the specified type.
        """
    
    @overload
    def element(self,
                __elmType: Literal["Truss"],
                tag: int,
                nodes: tuple,
                section: int,
                *args
            ) -> int:
        """
        Create a truss element.
        
        :param tag: unique :ref:`element` tag
        :type tag: |integer|
        :param nodes: tuple of *two* integer :ref:`node` tags
        :type nodes: tuple
        :param section: tag of a previously-defined :ref:`section`
        :type section: |integer|
        """

    @overload
    def element(self,
                __elmType: Literal["PrismFrame"],
                tag: int,
                nodes: tuple,
                section: int,
                transform: int,
                *args
            ) -> int:
        """
        Create a linear elastic prismatic frame element.        
        
        :param tag: unique :ref:`element` tag
        :type tag: |integer|
        :param nodes: tuple of *two* integer :ref:`node` tags
        :type nodes: tuple
        :param section: tag of a previously-defined :ref:`section`
        :type section: |integer|
        :param transform: identifier for previously-defined coordinate-transformation
        :type transform: |integer|
        """

    @overload
    def element(self,
                __elmType: Literal["ForceFrame"],
                tag: int,
                nodes: tuple,
                nip: int,
                transform: int,
                *args
               ) -> int: ...

class _Algorithm:
    def numberer(self, type: str, *args, **kwargs):
        """
        Create a numberer of the specified type.
        """

    def constraints(self, type: str, *args, **kwargs):
        """
        Create a constraint handler of the specified type.
        """

    def system(self, type: str, *args, **kwargs):
        """
        Create a system of the specified type.
        """

    def integrator(self, type: str, *args, **kwargs):
        """
        Create an integrator of the specified type.
        """
    
    def algorithm(self, type: str, *args, **kwargs) -> int:
        """
        Create an algorithm of the specified type.
        """

    def algorithm(self, 
                  type: Literal["NewtonLineSearch"],
                  tol=0.8) -> int:
        """
        Create a specific algorithm for the analysis.
        """

    def analysis(self, type: str):
        """
        Create an analysis of the specified type.
        """

    def analyze(self, nsteps: int = 1, *args) -> int:
        """
        Perform the analysis for a specified number of steps.
        
        :param nsteps: number of analysis steps to perform
        :type nsteps: |integer|
        """

    def test(self, type: str, *args, **kwargs):
        """
        Create a convergence test of the specified type.
        """

    def numIter(self) -> int:
        """
        Get the number of iterations for the last analysis step.
        """

    def wipe(self)->None:
        """
        Clear the current model.
        """


class _Output:
    def nodeDisp(self, node: int, dof: int) -> float:
        """
        Get the displacement of a node in a specified degree of freedom.
        """

    def getNodeTags(self) -> list[int]:
        """
        Get a list of all node tags in the model.
        """

    def nodeCoord(self, node: int, index: int=None) -> list[float] | float:
        """
        Get the coordinates of a node.
        """

    @overload
    def nodeDisp(self, node: int) -> list[float]:
        """
        Return the displacements of a node.
        """

    @overload
    def nodeDisp(self, node: int, dof: int) -> float:
        """
        Return the displacement of a node.
        """

    @overload
    def nodeReaction(self, node: int) -> list[float]:
        """
        Get the reaction forces at a node.
        """
    @overload
    def nodeReaction(self, node: int, dof: int) -> float:
        """
        Get the reaction force(s) at a node.
        """

    def getEleTags(self) -> list[int]:
        """
        Get a list of all element tags in the model.
        """

    def eleResponse(self, ele: int, responseType: str) -> list[float]:
        """
        Get the response of an element.
        """

class Model(_Materials,_Elements,_Algorithm, _Output):
    def node(self, tag: int, coords: tuple, *args, **kwargs) -> int:
        """
        Create a node with the specified tag and coordinates.
        """

    def fix(self, node: int, *args) -> int:
        """
        Apply boundary conditions to a node.
        """

    def mass(self, node: int, *args):
        """
        Assign mass to a node.
        """

    def load(self, node: int, *args):
        """
        Apply loads to a node.
        """



import warnings 
from itertools import cycle

import numpy as np
import matplotlib.pyplot as plt


from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

import numpy as np
import matplotlib.pyplot as plt


def _t_ppf(p: float, df: int) -> float:
    """
    Student-t inverse CDF (ppf). Only needed if CI is drawn.
    Uses SciPy if available; otherwise errors when CI is requested.
    """
    try:
        from scipy.stats import t as student_t  # type: ignore
    except Exception as e:
        raise ImportError(
            "Confidence intervals require SciPy (scipy.stats.t). "
            "Install with `pip install scipy`, or call draw(show_ci=False)."
        ) from e
    return float(student_t.ppf(p, df))


@dataclass
class _StepFit:
    x: float
    slope: float
    se_slope: float      # standard error of slope (sqrt(SSE) in MATLAB code)
    dfe: int             # degrees of freedom (n_pairs - 2)
    n_used: int          # number of error points used (after filtering/slicing)
    n_raw: int           # raw len(norms) for the step


class PlotConvergenceRate:
    """
    Convergence-rate plotter driven by xara.Model.testNorms().

    Assumes:
      - After each model.analyze(1), model.testNorms() returns the list
        of test norms for *that step's* Newton iterations.

    Computes (per step):
      log(e_{k+1}) = p * log(e_k) + c
    and stores p as the "convergence rate" (MATLAB slope).
    """

    def __init__(
        self,
        *,
        ci: Optional[float] = None,     # percent (e.g., 95). None/0 disables.
        n_ex_start: int = 1,            # ignore this many norms from start
        n_ex_end: int = 0,              # ignore this many norms from end
        n_total: int = 0,               # if >0, use only last n_total norms (>=3)
        step_start: float = 0.0,        # x-offset when x_mode="step"
        x_mode: str = "time",           # "step" or "time"
        x_label: Optional[str] = None,  # if None, chosen from x_mode
        y_label: str = "Convergence Rate",
        marker: str = None,
        markers = None,
        line_style: str = "none",       # MATLAB-like: markers only
        marker_size: float = 2.0,
        y_lim: Tuple[float, float] = (0.0, 5.0),
        y_ticks: Optional[Sequence[float]] = (0, 1, 2, 3, 4, 5),
        ci_alpha: float = 0.10,
        ax: Optional[plt.Axes] = None,
        skip: bool = False,
    ):
        self._skip = bool(skip)

        self.ci = None if (ci is None or ci == 0) else float(ci)
        self.n_ex_start = int(n_ex_start)
        self.n_ex_end = int(n_ex_end)
        self.n_total = int(n_total)
        self.step_start = float(step_start)
        self.x_mode = str(x_mode).strip().lower()

        if self.n_total != 0 and self.n_total < 3:
            raise ValueError("If n_total is used, it must be >= 3.")

        if self.x_mode not in ("step", "time"):
            raise ValueError('x_mode must be "step" or "time".')
        if markers is None and marker is None:
            markers = cycle(('.', 'o', 's', '^', 'v', 'x', '+'))
        elif marker is None:
            markers = cycle((marker,))

        self.markers = markers

        self.line_style = line_style
        self.marker_size = float(marker_size)
        self.y_lim = tuple(map(float, y_lim))
        self.y_ticks = None if y_ticks is None else list(map(float, y_ticks))
        self.ci_alpha = float(ci_alpha)

        # Figure/axes
        if not self._skip:
            if ax is None:
                fig, ax = plt.subplots()
                self.fig = fig
            else:
                self.fig = ax.figure
            self.ax = ax

            if x_label is None:
                x_label = "Load Step" if self.x_mode == "step" else "Load factor"
            self.ax.grid(True)
            self.ax.set_xlabel(x_label)
            self.ax.set_ylabel(y_label)
            self.ax.axhline(2, color="gray",  linestyle="-", linewidth=1)
            self.ax.axvline(0, color="black", linestyle="-", linewidth=1)
            self.ax.axhline(0, color="black", linestyle="-", linewidth=1)

        # Track global x-range over multiple series
        self._global_xmin = np.inf
        self._global_xmax = -np.inf

        self.reset(label="")

    def reset(self, label: str):
        """Start a new series (new line on the same axes)."""
        self._label = str(label)
        self._step_index = 0
        self._fits: List[_StepFit] = []
        self.marker = next(self.markers)

    def _step_x(self, model) -> float:
        """Choose x-coordinate for this step."""
        if self.x_mode == "time":
            # model.getTime() should exist in xara/opensees-style APIs
            return float(model.getTime())
        # "step" mode
        self._step_index += 1
        return self.step_start + self._step_index

    def _fit_slope_from_norms(self, norms: Sequence[float]) -> Tuple[float, float, int, int]:
        """
        Returns:
          slope, se_slope, dfe, n_used
        with NaNs / negative dfe if not computable.
        """
        e = np.asarray(list(norms), dtype=float).ravel()
        n_raw = int(e.size)

        if n_raw == 0:
            return np.nan, np.nan, -1, 0

        # Slice off start/end iterations (MATLAB: Punrm(1+NExStart:end-NExEnd))
        start = self.n_ex_start
        end = n_raw - self.n_ex_end if self.n_ex_end > 0 else n_raw
        if end <= start:
            return np.nan, np.nan, -1, 0

        e = e[start:end]

        # Ignore exactly 0 (MATLAB), and also guard against <=0 for log safety.
        e = e[np.isfinite(e)]
        e = e[e != 0.0]
        e = e[e > 0.0]

        n_used = int(e.size)

        # Need at least 3 error points to have >=2 pairs (MATLAB gate: NErr < 3 -> continue)
        if n_used < 3:
            return np.nan, np.nan, -1, n_used

        # If n_total is specified, use only the last n_total error points
        if self.n_total:
            if n_used < self.n_total:
                return np.nan, np.nan, -1, n_used
            e = e[-self.n_total :]
            n_used = int(e.size)

        # Regression pairs
        Xp = np.log(e[:-1])
        Yp = np.log(e[1:])

        devx = Xp - Xp.mean()
        devy = Yp - Yp.mean()
        sxx = float(np.sum(devx * devx))
        if sxx == 0.0:
            return np.nan, np.nan, -1, n_used

        slope = float(np.sum(devx * devy) / sxx)

        # Standard error of slope (for CI later)
        n_pairs = int(Xp.size)
        dfe = n_pairs - 2
        se_slope = np.nan
        if dfe > 0:
            intercept = float(Yp.mean() - slope * Xp.mean())
            resid = (Yp - intercept - slope * Xp)
            mse = float(np.sum(resid * resid) / dfe)
            sse = mse / sxx  # squared standard error of slope
            se_slope = float(np.sqrt(sse))

        return slope, se_slope, dfe, n_used

    def update(self, model):
        """
        Called once per load step (i.e., after each model.analyze(1) call).
        Reads norms from model.testNorms() and stores the fitted slope for that step.
        """
        if self._skip:
            return

        x = self._step_x(model)

        try:
            norms = model.testNorms()
        except Exception as e:
            raise AttributeError(
                "Model does not provide testNorms(), which is required for PlotConvergenceRate.update(model)."
            ) from e

        if norms is None:
            norms = []

        slope, se_slope, dfe, n_used = self._fit_slope_from_norms(norms)
        self._fits.append(
            _StepFit(
                x=x,
                slope=slope,
                se_slope=se_slope,
                dfe=dfe,
                n_used=n_used,
                n_raw=len(norms),
            )
        )

    def draw(self, *, show_ci: bool = False, **plot_kwargs):
        """
        Draw the current series.
        - Uses a new Matplotlib cycle color each time draw() is called.
        - If show_ci=True and ci is set, draws a translucent CI band.
        """
        if self._skip:
            return

        if not self._fits:
            warnings.warn("No steps recorded for this series.", stacklevel=2)
            return

        x = np.array([f.x for f in self._fits], dtype=float)
        y = np.array([f.slope for f in self._fits], dtype=float)

        have = np.isfinite(y)
        if not np.any(have):
            warnings.warn(
                "Not enough usable iterations in any step to determine convergence rate.",
                stacklevel=2,
            )
            return

        (line,) = self.ax.plot(
            x[have],
            y[have],
            label=self._label if self._label else None,
            linestyle=self.line_style,
            marker=self.marker,
            markersize=self.marker_size,
            **plot_kwargs,
        )
        color = line.get_color()

        # Optional CI (computed at draw time so SciPy is only required if you request CI)
        if show_ci and (self.ci is not None):
            lb = np.full_like(y, np.nan, dtype=float)
            ub = np.full_like(y, np.nan, dtype=float)

            alpha = 1.0 - self.ci / 100.0
            pcrit = 1.0 - alpha / 2.0

            for i, f in enumerate(self._fits):
                if not (np.isfinite(f.slope) and np.isfinite(f.se_slope) and f.dfe > 0):
                    continue
                t_score = _t_ppf(pcrit, f.dfe)
                me = t_score * f.se_slope
                lb[i] = f.slope - me
                ub[i] = f.slope + me

            have_ci = have & np.isfinite(lb) & np.isfinite(ub)
            if np.any(have_ci):
                self.ax.fill_between(
                    x[have_ci],
                    lb[have_ci],
                    ub[have_ci],
                    color=color,
                    alpha=self.ci_alpha,
                    linewidth=0,
                )

        # Update global x-range
        self._global_xmin = min(self._global_xmin, float(np.min(x[have])))
        self._global_xmax = max(self._global_xmax, float(np.max(x[have])))

    def finalize(self, *, title: Optional[str] = None):
        """Finalize formatting after all series are drawn."""
        if self._skip:
            return

        if title:
            self.ax.set_title(title)
        if len(self.ax.get_legend_handles_labels()):
            self.ax.legend()

        # MATLAB-ish y formatting
        self.ax.set_ylim(self.y_lim)
        if self.y_ticks is not None:
            self.ax.set_yticks(self.y_ticks)

        # Reasonable xlim across all series
        if np.isfinite(self._global_xmin) and np.isfinite(self._global_xmax):
            if self.x_mode == "step":
                # emulate MATLAB xlim([0 np] + StepStart) loosely
                self.ax.set_xlim(self.step_start, self._global_xmax + 1.0)
            else:
                self.ax.set_xlim(self._global_xmin, self._global_xmax)


    def savefig(self, filename: str, **savefig_kwargs):
        """Save the figure to a file."""
        if self._skip:
            return
        self.fig.savefig(filename, **savefig_kwargs)


#
#  FPS 
#
Length = "foot"
Force  = "lbf"
Mass   = "slug"


#
#  Length 
#
mm           = millimeter   = 0.00328084          # Milimeter
cm           = centimeter   = 0.032808399         # Centimeter
m            = meter        = 3.280839895         # Meter
km           = kilometer    = 3280.839895         # Kilometer
inch                        = 0.083333333         
ft           = foot         = 1.0                 # International foot
yd           = yard         = 3.0                 # International yard
mi           = mile         = 5280.0              # Mile

#
#  Angle 
#
rad          = radian       = 1.0                 # Radian
arcsec                      = 4.84814e-06         # Arcsecond
arcmin                      = 0.000290888         # Arcminute
deg          = degree       = 0.017453293         # Degrees

#
#  Constants 
#
pi           = pi           = 3.141592654         # Pi

#
#  Time 
#
sec          = second       = 1.0                 # second
min          = minute       = 60.0                # minute
hr           = hour         = 3600.0              # hour
day                         = 86400.0             # day
wk           = week         = 604800.0            # week
yr           = year         = 31536000.0          # year

#
#  Force 
#
N            = newton       = 0.224808941         # Newton (force)
dyn          = dyne         = 2.24809e-06         # Dyne
pdl          = poundal      = 0.03108095          # Poundal
lbf          = poundf       = 1.0                 
kip          = klbf         = 1000.0              

#
#  Mass 
#
slug                        = 1.0                 
lbm          = lbm          = 0.03108095          # International avoirdupois pound
gm           = gram         = 6.85218e-05         
kg           = kilogram     = 0.068521765         # Kilogram
tonne                       = 68.52176521         # Metric tonne
oz           = ounce        = 0.001942559         # International avoirdupois ounce

#
#  Pressure 
#
Pa           = pascal       = 0.020885434         # Pascal
torr                        = 2.784495531         
kPa          = kilopascal   = 20.88543404         
bar                         = 2088.543404         
atm          = atmosphere   = 2116.216604         # Standard atmosphere
MPa          = megapascal   = 20885.43404         
GPa          = gigapascal   = 20885434.04         
psi                         = 144.0               # Pound-square-inch
ksi                         = 144000.0

#
#  Density 
#
pcf                         = lbm / ft**3         # Pounds per cubic foot

#
#  Amount 
#
mol          = mole         = 1.0                 # Mole

#
#  Frequency 
#
Hz           = hertz        = 1.0                 

#
#  Area 
#
mm2                         = 1.07639e-05         # Square millimeter
m2                          = 0.001076391         # Square meter
cm2                         = 10.76391042         # Square centimeter
km2                         = 10763910.42         # Square kilometer
in2          = inch2        = 0.006944444         # Square inch
ft2          = feet2        = 1.0                 # Square foot
yd2          = yard2        = 9.0                 # Square yard
mi2          = mile2        = 27878400.0          # Square mile

#
#  Volume 
#
mm3                         = 3.53147e-08         
m3                          = 3.53147e-05         
cm3                         = 35.31466672         
km3                         = 35314666721.0       
in3          = inch3        = 0.000578704         
ft3          = foot3        = 1.0                 
cyd          = yard3        = 27.0                
mi3          = mile3        = 147198000000.0      

#
#  Velocity 
#
mmps                        = 0.00328084          # Millimeter per second
cps          = cmps         = 0.032808399         # Centimeter per second
mps                         = 3.280839895         # Meter per second
kps                         = 3280.839895         # Kilometer per second
ips          = inchps       = 0.083333333         # Inch per second
fps          = footps       = 1.0                 # Foot per second
yps                         = 3.0                 # Yard per second
mph                         = 1.466666667         # mile per hour

#
#  Acceleration 
#
mmps2                       = 0.00328084          
cps2         = cmps2        = 0.032808399         
mps2                        = 3.280839895         
kps2                        = 3280.839895         
ips2         = inchps2      = 0.083333333         
fps2         = footps2      = 1.0                 
yps2                        = 3.0                 
mph2                        = 0.000407407         
gravity                     = 32.17404856         # Standard gravity

#
#  Angular Velocity 
#
rpm          = revpm        = 6.283185307         # Revolution per minute
radps                       = 1.0                 # Radian per second



import xara
from collections import defaultdict

class _Parameter(float):
    __slots__ = ("index", "name")

    def __new__(cls, value, index=0, name=None):
        obj = super().__new__(cls, value)
        obj.index = int(index)
        obj.name  = str(name)
        return obj

    def __repr__(self):
        return f"{{Parameter {self.index} {super().__repr__()}}}"



class ParaModel:
    """
    A wrapper around xara.Model that allows us to use JAX's vjp/jvp
    with the model's methods.
    """
    def __init__(self, *args, parameters=0, **kwargs):
        self._model = xara.Model(*args, **kwargs)
        for i in range(parameters):
            self._model.parameter(i)

        self._p = defaultdict(lambda: {
            "material": set(),
            "section":  set(),
            "element":  set()
        })


    def __getattr__(self, name):
        return getattr(self._model, name)


    def analyze(self, *args, **kwargs):
        self.sensitivityAlgorithm("-computeAtEachStep")
        for (index,name),param in self._p.items():
            for level in "element",:
                for tag in param[level]:
                    self._model.addToParameter(index, level, tag, name)

        return self._model.analyze(*args, **kwargs)


    def _pull_parameters(self, level, tag, args, kwds):
        new_args = []
        new_kwds = {}
        for arg in args:
            if isinstance(arg, _Parameter):
                self._p[(arg.index,arg.name)][level].add(tag)
                new_args.append(float(arg))
            else:
                new_args.append(arg)

        for key, arg in kwds.items():
            if isinstance(arg, _Parameter):
                self._p[(arg.index,arg.name)][level].add(tag)
                new_kwds[key] = float(arg)
            else:
                new_kwds[key] = arg

        if "material" in kwds:
            for param in self._p.values():
                if kwds["material"] in param["material"]:
                    param[level].add(tag)

        if "section" in kwds:
            for param in self._p.values():
                if kwds["section"] in param["section"]:
                    param[level].add(tag)

        return new_args, new_kwds

    def section(self, *args, **kwds):
        tag = args[1]
        args, kwds = self._pull_parameters("section", tag, args, kwds)
        return self._model.section(*args, **kwds)

    def material(self, *args, **kwds):
        tag = args[1]
        args, kwds = self._pull_parameters("material", tag, args, kwds)
        return self._model.material(*args, **kwds)
    
    def element(self, *args, **kwds):
        tag = args[1]
        args, kwds = self._pull_parameters("element", tag, args, kwds)
        return self._model.element(*args, **kwds)

try:
    from jax import tree_util
    def _flatten_model(model):
        return (), model

    def _unflatten_model(aux_data, _children):
        return aux_data

    tree_util.register_pytree_node(ParaModel, _flatten_model, _unflatten_model)
except ImportError:
    # JAX is not available, so we won't register the pytree node
    pass



import xara 

elem = 1
node = 1

options = {
    "integrator": {
        "type": "LoadControl",
        "step": 10,
    }
}

model = xara.Model(ndm=3, ndf=6)

analysis = xara.StaticAnalysis(model, loads, options)


for step in analysis.steps():
    state = xara.solve(model, step)
    if state.status != xara.successful:
        raise RuntimeError("Analysis did not complete successfully.")

    state.node(node).solution(dof=None)
    state.node(node).rotation(dof=None)
    state.node(node).reaction(dof=None)
    state.cell(elem).response(key=None, **kwds)
    state.iter.size
    step.size


for state in map(xara.solve, analysis.steps()):
    
    step.time
    state.node(node).solution(dof=None)
    step.node(node).rotation(dof=None)
    step.node(node).reaction(dof=None)
    step.cell(elem).response(key=None, **kwds)
    step.iter.size
    step.size






from xsection.library import * 


class ShellShape:
    def __init__(self, thickness, group=None, n: int = None):
        self.thickness = thickness
        self.group = group
        self.n = n



class CompositeShell:
    def __init__(self, shapes):
        pass




from pathlib import Path
import math

try:
    import orjson as json

except ImportError:
    import json


__all__ = ["UnitHandler"]

JSON_DIR = Path(__file__).parents[0] / "json/"


def load(system, overrides=(), abbrev=True):
    abbrev_ext = "-w_abbrev" if abbrev else ""

    def_file = JSON_DIR / f"{system}{abbrev_ext}.json"
    try:
        with open(def_file, "r") as f:
            defs = json.load(f)

    except FileNotFoundError:
        raise ValueError(f"Cannot find unit definition file {def_file}")

    return defs


class SpacedDimensions:
    def __init__(self, step, justification):
        self.justification = justification
        self.step = step

    def __matmul__(self, num):
        assert isinstance(num, int)
        if self.justification in ["c", "centered"]:
            begin = -sum([self.step] * num) / 2

        elif self.justification in ["l", "left"]:
            begin = 0.0

        return [begin + self.step * i for i in range(num + 1)]

    def __rmatmul__(self, num):
        return self.__matmul__(num)


class Dimension(float):
    def __mul__(self, other):
        try:
            out = float.__mul__(self, float(other))
        except:
            out = type(other)(
                [
                    # float mult if scalar, else recurse
                    float.__mul__(self, i) if isinstance(i, (float, int)) else other * i
                    for i in other
                ]
            )
        return out

    def __rmul__(self, other):
        return self * other


class UnitHandler:
    def __init__(self, base, abbrev=True):
        system, *overrides = base.split(", ")
        self.base_dims = load(system, overrides)

    def __getattr__(self, name):
        return Dimension(self.base_dims[name])

    def __repr__(self):
        return f"""<{', '.join(self.base_dims.keys())}>"""

    def dim(self, dim):
        feet, inches = map(float, dim.split("-"))
        return feet * self.foot + inches * self.inch

    def spacing(self, dim, justification="l"):
        if isinstance(dim, str):
            dim = self.dim(dim)
        return SpacedDimensions(dim, justification)

def spacing(dim, just):
    return SpacedDimensions(dim, just)


_Aliases = {
    "mm": "millimeter",
    "cm": "centimeter",
    "m": "meter",
    "km": "kilometer",
    "ft": "foot",
    "yd": "yard",
    "mi": "mile",

    "N": "newton",
    "dyn": "dyne",
    "pdl": "poundal",
    "kip": "kilopoundf",
    "lbf": "poundf",

    "lbm": "poundmass",
    "gm": "gram",
    "kg": "kilogram",
    "oz": "ounce",

    "Pa": "pascal",
    "torr": "torr",
    "kPa": "kilopascal",
    "bar": "bar",
    "atm": "atmosphere",
    "MPa": "megapascal",
    "GPa": "gigapascal",

    "Hz": "hertz",

    "rad": "radian",
    "deg": "degree",

    "sec": "second",
    "min": "minute",
    "hr": "hour",
    "wk": "week",
    "yr": "year",
}

class UnitSystem:
    #
    #  IPS 
    #
    Length: str
    Force : str
    Mass  : str

    _length = {
        "inch", "foot", "yard", "mile",
        "millimeter", "centimeter", "meter", "kilometer"
    }
    _force = {
        "poundf", "kilopoundf",
        "newton", "dyne", "poundal"
    }
    _mass = {
        "poundmass", "slug", ""
        "gram", "kilogram", "ounce"
    }


    #
    #  Length 
    #
    millimeter   : float         # Milimeter
    centimeter   : float         # Centimeter
    meter        : float         # Meter
    kilometer    : float         # Kilometer
    inch         : float
    foot         : float         # International foot
    yard         : float         # International yard
    mile         : float         # Mile

    #
    #  Angle 
    #
    radian       : float                 # Radian
    arcsec                      : float         # Arcsecond
    arcmin                      : float         # Arcminute
    degree       : float         # Degrees

    #
    #  Constants 
    #
    pi           : float   = math.pi      # Pi

    #
    #  Time 
    #
    second       : float
    minute       : float
    hour         : float
    day          : float
    week         : float
    year         : float

    #
    #  Force 
    #
    newton       : float         # Newton (force)
    dyne         : float         # Dyne
    poundal      : float         # Poundal
    poundf       : float
    kilopoundf   : float

    #
    #  Mass 
    #
    slug         : float
    poundmass    : float         # International avoirdupois pound
    gram         : float
    kilogram     : float         # Kilogram
    tonne        : float         # Metric tonne
    ounce        : float         # International avoirdupois ounce

    #
    #  Pressure 
    #
    pascal       : float         # Pascal
    torr         : float         
    kilopascal   : float         
    bar          : float         
    atmosphere   : float         # Standard atmosphere
    megapascal   : float         
    gigapascal   : float         
    psi          : float         # Pound-square-inch
    ksi          : float

    #
    #  Frequency
    #
    hertz        : float                 

    #
    #  Area 
    #
    mm2          : float         # Square millimeter
    m2           : float          # Square meter
    cm2          : float           # Square centimeter
    km2          : float        # Square kilometer
    in2          : float                 # Square inch
    ft2          : float               # Square foot
    yd2          : float              # Square yard
    mi2          : float        # Square mile

    #
    #  Volume 
    #
    mm3          : float         
    m3           : float         
    cm3          : float         
    km3          : float    
    in3          : float                 
    ft3          : float              
    cyd          : float
    mi3          : float

    #
    #  Velocity 
    #
    mmps         : float         # Millimeter per second
    cps          : float         # Centimeter per second
    mps          : float         # Meter per second
    kps          : float         # Kilometer per second
    ips          : float                # Inch per second
    fps          : float                # Foot per second
    yps          : float                # Yard per second
    mph          : float                # mile per hour

    #
    #  Acceleration 
    #
    mmps2      : float
    cps2       : float         
    mps2       : float         
    kps2       : float         
    ips2       : float                 
    fps2       : float                
    yps2       : float                
    mph2       : float         
    gravity    : float         # Standard gravity

    #
    #  Angular Velocity 
    #
    rpm        : float         # Revolution per minute
    radps      : float         # Radian per second





if __name__ == "__main__":
    import sys
    import xara

    model = xara.Model()
    loads = [xara.NodalLoad(node=1, dof=1, value=1.0)]
    trace = xara.Trace(model, loads, method=xara.ArcLengthControl())

    while model.nodeDisp(1,1) < 0.1:
        trace.step()
        print(model.nodeDisp(1,1))


import sys
from . import units
HELP = """
python -m libunit <dst>/<frm>...
                  <sys> <dst>...

Examples

python -m libunit si inch foot pound

python -m libunit m/inch newton/pound
"""
def parse_args(args):
    config = {"a": None, "b": None}
    argi = iter(args[1:])
    for arg in argi:
        if False:
            pass
        if config["a"] is None:
            config["a"] = arg
        elif config["b"] is None:
            config["b"] = arg
    return config

if __name__ == "__main__":
    config = parse_args(sys.argv)
    print(getattr(units.UnitHandler(config["a"]), config["b"]))




_UniaxalMaterials = [
    "Steel02",
    "Steel01",
    "Concrete01",
    "Concrete02",
    "Pinching4",
    "ASDConcrete1D"
]


def Material(type=None, name=None, *args, **kwds):
    if type is None:
        if len(kwds) <= 2:
            type = "Elastic"
        else:
            raise ValueError("Material type must be specified when more than 2 parameters are provided.")

    if type == "Elastic":
        return _ElasticMaterial(*args, **kwds)
    elif type in _UniaxalMaterials:
        return _UniaxialMaterial(type, *args, **kwds)
    elif "J2" in type or type == "Plastic":
        return _PlasticMaterial(type, *args, **kwds)
    else:
        return _MultiaxialMaterial(type, *args, **kwds)



class _Material:
    def __init__(self, type, tag=None, **kwds):
        self.type = type
        self._mtag = tag

    def _add_to_model(self, model, tag):
        raise NotImplementedError("Material type must implement '_add_to_model' method.")

    def __getitem__(self, key):
        return getattr(self, key)


class _UniaxialMaterial(_Material):
    def __init__(self, type, *args, G=None, nu=None, **kwds):
        super().__init__(type, **kwds)
        self._args = args
        kwds = {self._normalize_parameter(k): v for k, v in kwds.items()}
        self._kwds = kwds
        if "E" in kwds:
            self.E = kwds["E"]
        
        if type in {"Concrete01", "Concrete02"}:
            if "E" not in kwds:
                if "Fc" not in kwds or "ec0" not in kwds:
                    raise ValueError("Concrete materials require either E or both Fc and ec0.")
                self.E = 2*kwds["Fc"]/kwds["ec0"]

        if G is not None and nu is not None:
            raise ValueError("Only one of Poisson's ratio (nu) or shear modulus (G) should be provided.")
        elif G is not None:
            self.G = G
            self.nu = self.E / (2 * G) - 1
        elif nu is not None:
            self.G = self.E / (2 * (1 + nu))
            self.nu = nu

    def _add_to_model(self, model, tag):
        self._mtag = tag
        kwds = {k: v for k, v in self._kwds.items() if k not in {"G", "nu"}}
        if self.type in {"Concrete01", "Concrete02"} and "E" in kwds:
            del kwds["E"]
        model._invoke_proc("uniaxialMaterial", self.type, tag, **kwds)

    def _normalize_parameter(self, name):
        if self.type in {"Concrete01", "Concrete02"}:
            if name in {"ec0", "epsc0"}:
                return "ec0"
            elif name in {"fc", "Fc"}:
                return "Fc"
        return name
        


class _ElasticMaterial(_Material):
    def __init__(self, E, nu=None, G=None, tag=None):
        super().__init__("Elastic", tag=tag)
        self.E = E
        if nu is None and G is None:
            raise ValueError("Either Poisson's ratio (nu) or shear modulus (G) must be provided.")
        elif nu is not None and G is not None:
            raise ValueError("Only one of Poisson's ratio (nu) or shear modulus (G) should be provided.")
        elif nu is not None:
            self.nu = nu
            self.G = E / (2 * (1 + nu))
        else:
            self.G = G
            self.nu = E / (2 * G) - 1
    
    def _add_to_model(self, model, tag):
        self._mtag = tag
        model.material("ElasticIsotropic", tag, E=self.E, G=self.G)


class _MultiaxialMaterial(_Material):
    def __init__(self, type, *args, **kwds):
        super().__init__(type)
        self._args = args
        self._kwds = kwds
        E = self.E = kwds.get("E", None)
        G = kwds.get("G", None)
        nu = kwds.get("nu", None)

        if nu is None and G is None:
            raise ValueError("Either Poisson's ratio (nu) or shear modulus (G) must be provided.")
        elif nu is not None and G is not None:
            raise ValueError("Only one of Poisson's ratio (nu) or shear modulus (G) should be provided.")
        elif nu is not None:
            self._v = nu
            self.G = E / (2 * (1 + nu))
        else:
            self.G  = G
            self._v = E / (2 * G) - 1

    def _add_to_model(self, model, tag):
        self._mtag = tag
        model._invoke_proc("material", self.type, tag, *self._args, **self._kwds)

class _PlasticMaterial(_Material):
    def __init__(self, type, E, nu=None, G=None, Fy=None, Hiso=None, Hkin=None, **kwds):
        super().__init__(type)
        self.E = E
        self.Fy = Fy
        self.Hiso = Hiso
        self.Hkin = Hkin

        if nu is None and G is None:
            raise ValueError("Either Poisson's ratio (nu) or shear modulus (G) must be provided.")
        elif nu is not None and G is not None:
            raise ValueError("Only one of Poisson's ratio (nu) or shear modulus (G) should be provided.")
        elif nu is not None:
            self._v = nu
            self.G = E / (2 * (1 + nu))
        else:
            self.G = G
            self._v = E / (2 * G) - 1

    
    def _add_to_model(self, model, tag):
        self._mtag = tag
        mdata = {
            "E": self.E,
            "G": self.G,
            "Fy": self.Fy,
        }
        for key in ["Hiso", "Hkin"]:
            value = getattr(self, key)
            if value is not None:
                mdata[key] = value

        model.material(self.type, tag, **mdata)



from pathlib import Path

# An example lives in a file called "model.py" under a directory named after the example, e.g. "cantilever/model.py". 

# The file model.py should define the following:
# - a function called "create_model" that returns a xara.Model object. 
# - a function called "solve" whose first argument is a xara.Model, and the remaining arguments are the options declared in solve_options. 
#   
# The file test.py must define at least the following:
# - a variable called model_options that is an instance of the Options class defined and 
#   declares the arguments to the create_model function.
# - a variable called solve_options that is an instance of the Options class defined and 
#   declares the additional arguments to the solve function.

# Options is a container of Option objects. an Option must be convertible to 
# args,kwds that can be passed to the create_model and solve functions.

# model.py should be simple and readable by beginners. 
# test.py can be more complex and is for internal validation purposes

# An example lives in a file called "model.py" under a directory named after
# the example, e.g. "cantilever/model.py".
#
# The file model.py should define the following:
#   - a function called "create_model" that returns a xara.Model object.
#   - a function called "solve" whose first argument is a xara.Model, and
#     the remaining arguments are the options declared in solve_options.
#
# model.py should be simple and readable by beginners.
#
# The file test.py must define at least the following:
#   - a variable called model_options (Options instance) declaring the
#     arguments to create_model.
#   - a variable called solve_options (Options instance) declaring the
#     additional arguments to solve.
#
# test.py can be more complex and is for internal validation purposes.

"""
xara.examples
=============

An example lives in a directory named after the example (e.g.
``frame-0001/``).  The directory must contain at least:

  model.py
      Defines ``create_model(**kwds)`` and ``solve(model, **kwds)``.
      May import from other files in the same directory.
      Should be simple and readable by beginners.

  test.py
      Defines ``Cases``: a list of dicts whose keys are arguments to
      ``create_model`` and/or ``solve``.  The Example object inspects
      function signatures to route each key to the right function.

      May also contain ``test_*()`` functions for pytest.
"""

from pathlib import Path
import importlib.util
import inspect
import sys

EXAMPLE_ROOT = Path("~/online/benchmarks/benchmarks").expanduser()


# ------------------------------------------------------------------
# Import helper
# ------------------------------------------------------------------

def _import_module_from_file(filepath: Path, module_name: str | None = None):
    """Import a .py file as a module.

    Adds the file's parent directory to ``sys.path`` so that sibling
    imports (e.g. ``from helpers import ...`` inside model.py) work,
    even when the directory name is not a valid Python identifier.
    """
    filepath = Path(filepath).resolve()

    # Ensure sibling imports work
    parent = str(filepath.parent)
    if parent not in sys.path:
        sys.path.insert(0, parent)

    if module_name is None:
        safe_parent = filepath.parent.name.replace("-", "_").replace(" ", "_")
        module_name = f"_xara_example_.{safe_parent}.{filepath.stem}"

    spec = importlib.util.spec_from_file_location(module_name, filepath)
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot import {filepath}")
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


# ------------------------------------------------------------------
# Example
# ------------------------------------------------------------------

class Example:
    """A loaded example.

    Attributes
    ----------
    name : str
        Directory name (e.g. ``"frame-0001"``).
    path : Path
        Absolute path to the example directory.
    model : module
        The imported ``model.py``.
    cases : list[dict]
        The ``Cases`` list from ``test.py``.
    """

    def __init__(self, name, path, model_module, test_module):
        self.name = name
        self.path = Path(path)
        self.model = model_module
        self.test = test_module
        self.cases: list[dict] = getattr(test_module, "Cases", [{}])

        # Cache the parameter names so we can split case dicts
        self._model_params = set(inspect.signature(self.model.create_model).parameters)
        self._solve_params = set(inspect.signature(self.model.solve).parameters) - {"model"}

    def __repr__(self):
        return f"Example({self.name!r}, {len(self.cases)} cases)"

    def _split(self, case: dict):
        """Split a flat case dict into (model_kwds, solve_kwds)."""
        model_kwds = {}
        solve_kwds = {}
        for k, v in case.items():
            if k in self._model_params:
                model_kwds[k] = v
            elif k in self._solve_params:
                solve_kwds[k] = v
            else:
                raise KeyError(
                    f"Case key {k!r} not found in create_model{sorted(self._model_params)} "
                    f"or solve{sorted(self._solve_params)}"
                )
        return model_kwds, solve_kwds

    def create_model(self, case: dict | None = None):
        model_kwds, _ = self._split(case or {})
        return self.model.create_model(**model_kwds)

    def solve(self, model, case: dict | None = None):
        _, solve_kwds = self._split(case or {})
        return self.model.solve(model, **solve_kwds)

    def run(self, case: dict | None = None):
        """``create_model`` then ``solve``.  Returns ``(model, result)``."""
        case = case or {}
        model_kwds, solve_kwds = self._split(case)
        model = self.model.create_model(**model_kwds)
        result = self.model.solve(model, **solve_kwds)
        return model, result


def load(name: str, root: Path | str | None = None) -> Example:
    """Load an example by directory name.

    Parameters
    ----------
    name : str
        Example directory name, e.g. ``"frame-0001"``.
    root : Path, optional
        Override ``EXAMPLE_ROOT``.
    """
    root = Path(root) if root is not None else EXAMPLE_ROOT
    example_dir = root / name

    model_file = example_dir / "model.py"
    test_file  = example_dir / "test.py"

    if not model_file.exists():
        raise FileNotFoundError(f"No model.py at {model_file}")
    if not test_file.exists():
        raise FileNotFoundError(f"No test.py at {test_file}")

    model_module = _import_module_from_file(model_file)
    test_module  = _import_module_from_file(test_file)

    return Example(name, example_dir, model_module, test_module)


import numpy as np

class Column:
    def __init__(self, E, I, L, H, Pr, w=None):
        self.E = E
        self.I = I
        self.L = L
        self.H = H
        self.Pr = Pr
        self.w  = w

    def create_model():
        pass


class Case1:
    # Cantilever with tip loads
    def __init__(self, E, I, L, H, Pr):
        Pe = (np.pi**2 * E * I) / L**2
        self.Pe = Pe
        lam = (np.pi / 2) * np.sqrt(Pr / Pe)
        self.u = (H * L**3 / (3 * E * I)) * (3 * (np.tan(2 * lam) - 2 * lam) / (2 * lam)**3)
        self.M = H * L * (np.tan(2 * lam) / (2 * lam))
        self.xu = L

class Case2:
    # Guided
    def __init__(self, E, I, L, H, Pr):
        lam = (np.pi / 2) * np.sqrt(Pr / ((np.pi**2 * E * I) / L**2))
        self.u = (H * L**3 / (12 * E * I)) * (3 * (np.tan(lam) - lam) / lam**3)
        self.M = (H * L / 2) * (np.tan(lam) / lam)
        self.xu = L

class Case3:
    def __init__(self, E, I, L, Pr, w):
        lam = (np.pi / 2) * np.sqrt(Pr / ((np.pi**2 * E * I) / L**2))
        self.u = (5 * w * L**4 / (384 * E * I)) * (12 * (2 / np.cos(lam) - lam**2 - 2) / (5 * lam**4))
        self.M = (w * L**2 / 8) * (2 * (1 / np.cos(lam) - 1) / lam**2)
        self.xu = L/2

class Case4:
    def __init__(self, E, I, L, Pr, w):
        lam = (np.pi / 2) * np.sqrt(Pr / ((np.pi**2 * E * I) / L**2))

        num = 12 * (2 - 2 * np.cos(lam) - lam * np.sin(lam))
        den = lam**3 * np.sin(lam)
        self.u = (w * L**4 / (384 * E * I)) * (num / den)

        self.M = (w * L**2 / 12) * ((3 * (np.tan(lam) - lam)) / (lam**2 * np.tan(lam)))
        self.xu = L/2

class Case5:
    # Propped cantilever
    def __init__(self, E, I, L, Pr, w):
        lam = (np.pi / 2) * np.sqrt(Pr / ((np.pi**2 * E * I) / L**2))
        sec_lam = 1 / np.cos(lam)
        tan_lam = np.tan(lam)
        tan_2lam = np.tan(2 * lam)

        den_common = (1 / (2 * lam)) - (1 / tan_2lam)
        num_inner = (tan_lam - lam) * (sec_lam - 1)

        inner_term = (2 * sec_lam - 2 * lam - 2) - (num_inner / den_common)
        self.u = (w * L**4 / (192 * E * I)) * (6 / lam**4) * inner_term

        self.M = (w * L**2 / 8) * (2 * (tan_lam - lam) / (lam**2 * den_common))
        self.xu = L/2
    
    def _fix(self, model, i, j, ndm):
        model.fix(i, ())


if __name__ == "__main__":
    from xara.units.iks import inch, ksi, ft, kip
    A =  14.1
    I =  484.0
    d =  13.8
    tw = 0.340
    section = dict(
            E  = 29000.0,
            I = 484.0*inch**4
    )

    c2 = Case1(**section,
              L=28*ft, 
              H=1*kip,
              Pr=150*kip)

    print("M: ", c2.M)
    print("u: ", c2.u)
